# small script to find phase cycles
import warnings
import numpy as np
import numba as nb

__version__ = '0.4'



#######################################################################
#                                                                     #
#   U   U  TTTTT   III   L       III   TTTTT   III   EEEEE   SSSS     #
#   U   U    T      I    L        I      T      I    E      S         #
#   U   U    T      I    L        I      T      I    EEE     SSS      #
#   U   U    T      I    L        I      T      I    E          S     #
#    UUU     T     III   LLLLL   III     T     III   EEEEE  SSSS      #
#                                                                     #
#######################################################################



def all_combis(*args):
    """
    Return all combinations obtained by choosing one element from N lists.
    
    An array is constructed containing each possible ordered combination
    obtained by choosing one element from a list of variable length per
    position. Each position has its own list.
    
    Parameters
    ----------
    args : array_like
        Lists containing elements for each position.
            
    Returns
    -------
    combis : 2D array
        Array of all combinations obtained by choosing one element from
        each list.
    """
    
    # initialize the lists
    elements = [np.asarray(e).flatten() for e in args]
    dtype = np.find_common_type([e.dtype for e in elements], [])
    
    # get length of list and products for number of repetitions
    lengths = [e.size for e in elements]
    cumprod = np.concatenate(([1], np.cumprod(lengths)))
    cumprod_inv = np.concatenate(([1], np.cumprod(lengths[::-1])))
    
    # allocate space
    combis = np.zeros(shape=(cumprod[-1], len(lengths)), dtype=dtype)
    
    for i in range(len(lengths)):
        combis[:,i] = np.tile(np.repeat(elements[i], cumprod_inv[-2-i]),
                              cumprod[i])
    
    return combis



def construct_ctps(nblocks, pmax, start=0, end=-1):
    """
    Construct an array with all possible coherence order transfer pathways
    possible.
    
    Parameters
    ----------
    nblocks : int
        Number of steps or pulse blocks changing the coherence order.
    pmax : (int, list of lists)
        Maximum possible coherence order present between all pulse blocks or
        a list of maximum possible coherence orders between each consecutive
        pair of pulse blocks.
    start : (int, array_like, str), optional
        Coherence order or list of possible coherence orders at the start.
        Defaults to 0. If 'all', set to pmax.
    end : (int, array_like, str), optional
        Coherence order of list of possible coherence orders at the end.
        Defaults to -1 assuming a perfect quadrature detector.
        If 'all', set to pmax.
    
    Returns
    -------
    CTPs : array_like
        All CTPs possible.
    """
    
    if type(pmax) is int:
    
        if start == 'all' : start = pmax
        if end   == 'all' :   end = pmax
        
        coherences = np.linspace(-pmax, pmax, 2*pmax+1, dtype=int)
        return all_combis(start, *[coherences]*(nblocks-1), end)
    
    elif hasattr(pmax, '__iter__'):
        
        # convert to array
        pmax = np.asarray(pmax).flatten()
        if not len(pmax) == nblocks-1:
            raise ValueError('Length of "pmax" must be one element shorter than the number of blocks.')
        
        coherences = [np.linspace(-p, p, 2*p+1, dtype=int) for p in pmax]
        return all_combis(start, *coherences, end)
        
    else:
        raise TypeError('"pmax" must be integer or a iterable list.')



def get_next_combis(start, steps, nblocks, divisor):
    """
    Generate combinations of integers in the half-open interval [0,divisor).
    
    This function effectively converts a range from consecutive integers
    from 'start' to 'start'+'steps'-1 into their representation in the base of 
    the divisor. This functions is useful to generate some combination of 
    winding numbers for exhaustive cogwheel phase cycle searches.
    
    Parameters
    ----------
    start : int
        Start of the array of consecutive integers that are to be converted.
    steps : int
        Number of steps and total number of integers to be converted.
    nblocks : int
        Number of blocks of the output array. Corresponds to the number of
        digits of the base-'divisor' number.
    divisor : int
        Divisor that corresponds to the base of counting. All numbers in 'out'
        will be elements of the half-open interval [0,divisor).
    
    Returns
    -------
    out : ndarray, shape (steps, nblocks)
        Output array with the first dimension encoding the converted number and
        the second dimension encoding the digit.
    """
    
    # some checks on the input
    if start < 0:
        raise ValueError("'start' must be non-negative!")
    if steps < 1:
        raise ValueError("There must be at least one step!")
    if nblocks < 1:
        raise ValueError("There must be a least one block!")
    if divisor < 1:
        raise ValueError("Divisor must be at least one!")

    # counting array, choose unsigned integers to represent large numbers    
    cnt = np.arange(start, start+steps, 1, dtype=np.uint64)
    
    # allocate memory for output array
    out = np.zeros(shape=(steps, nblocks), dtype=int)
    
    # start counting
    for i in range(nblocks):
        cnt, out[:,-1-i] = np.divmod(cnt, divisor, dtype=np.uint64)
        
    return out



def get_wantedctp_idx(ctps, ctps_wanted):
    """
    Get indices of desired pathways.

    Parameters
    ----------
    ctps : (nPaths, nBlocks), 2D array
        Coherence transfer paths (CTPs).
    ctps_wanted : (nWanted, nBlocks), 2D array
        List of wanted CTPs. If any CTP is given in 'ctps_wanted' which
        is not present in the 'ctps', a warning is given.
    """
    
    # convert input to arrays
    ctps = np.atleast_2d(np.asarray(ctps))
    ctps_wanted = np.atleast_2d(np.asarray(ctps_wanted))
    
    # check shape of input
    if not ctps.ndim == 2:
        raise ValueError("'ctps' must be two dimensional!")
    if not ctps_wanted.ndim == 2:
        raise ValueError("'ctps_wanted' must be one dimensonal!")
    if not ctps.shape[1] == ctps_wanted.shape[1]:
        raise ValueError("Sizes of 'ctps' and 'ctps_wanted' do not match!")
    
    allidx = np.linspace(0, ctps.shape[0]-1, ctps.shape[0], dtype=np.int64)
    mask = np.any(np.all(ctps[None,:,:] == ctps_wanted[:,None,:], axis=2), axis=0)
    
    return allidx[mask]



def split_ctps_wanted(ctps, ctps_wanted):
    """
    Split a list of CTPs in wanted and unwated CTPs.py
    
    Parameters
    ----------
    ctps : (nPaths, nBlocks), 2D array
        Coherence transfer paths (CTPs).
    ctps_wanted : (nWanted, nBlocks), 2D array
        List of wanted CTPs. If any CTP is given in 'ctps_wanted' which
        is not present in the 'ctps', a warning is given.

    Returns
    -------
    wanted_ctps : (nWanted, nBlocks)
        Wanted CTPs.
    unwanted_ctps : (nUnwanted, nBlocks)
        Unwanted CTPs.
    """

    # convert input to arrays
    ctps = np.atleast_2d(np.asarray(ctps))
    ctps_wanted = np.atleast_2d(np.asarray(ctps_wanted))
    
    # check shape of input
    if not ctps.ndim == 2:
        raise ValueError("'ctps' must be two dimensional!")
    if not ctps_wanted.ndim == 2:
        raise ValueError("'ctps_wanted' must be two dimensonal!")
    if not ctps.shape[1] == ctps_wanted.shape[1]:
        raise ValueError("Sizes of 'ctps' and 'ctps_wanted' do not match!")

    # split in wanted and unwated CTPs
    wanted_idx = get_wantedctp_idx(ctps, ctps_wanted)

    if wanted_idx.size < ctps_wanted.shape[0]:
        warnings.warn("DAS IST EIN TEST")

    wanted_ctps = np.copy(ctps[wanted_idx,:])
    unwanted_ctps = np.delete(ctps, wanted_idx, axis=0)
    
    return wanted_ctps, unwanted_ctps


def prepare_ctps(ctps, ctps_wanted):
    """Prepare ctps.
    
    Wanted ctps are copied to the beginning of the array.
    
    This function is compiled with numba.
    
    Parameters
    ----------
    ctps : ()
        pass
    ctps_wanted : ()
        pass
    
    Returns
    -------
    ctps_out : 
        pass
    """
    
    ctps = np.atleast_2d(ctps)
    ctps_wanted = np.atleast_2d(ctps_wanted)
    
    if not ctps.ndim == ctps_wanted.ndim == 2:
        raise ValueError("`ctps` and `ctps_wanted` must both be one or two dimensional.")
    if not ctps.shape[1] == ctps_wanted.shape[1]:
        raise ValueError("`ctps` and `ctps_wanted` must have equal number of columns.")

    ctps_out = np.zeros_like(ctps)
    _prepare_ctps_numba(ctps, ctps_wanted, ctps_out)

    # code w/o numba
    # idx = np.argwhere(np.any(np.all(ctps[:,None,:] == ctps_wanted[None,:,:], axis=2), axis=1))[:,0]
    # if not idx.size == ctps_wanted.shape[0]:
    #     raise ValueError("lol")
    
    # ctps_out = np.concatenate((ctps[idx,:], np.delete(ctps, idx, axis=0)), axis=0)
    
    return ctps_out


@nb.njit()
def _prepare_ctps_numba(ctps, ctps_wanted, ctps_out):
    """Prepare ctps.
    
    Wanted ctps are copied to the beginning of the array.
    
    This function is compiled with numba.
    
    Parameters
    ----------
    ctps : ()
        pass
    ctps_wanted : ()
        pass
    ctps_out : ()
        pass
    
    Returns
    -------
    ctps_out : 
        pass
    """
    

    # check shape and size of input
    if not ctps.ndim == 2:
        raise ValueError("Error in `_prepare_ctps_numba`: `ctps` must be castable to 2D.")
    if not ctps_out.ndim == 2:
        raise ValueError("Error in `_prepare_ctps_numba`: `ctps_out` must be castable to 2D.")
    if not ctps_wanted.ndim == 2:
        raise ValueError("Error in `_prepare_ctps_numba`: `ctps_wanted` must be castable to 2D.")
        
    n_ctps   = ctps.shape[0]
    n_blocks = ctps.shape[1]
    n_wanted = ctps_wanted.shape[0]

    if not ctps_out.shape[0] == n_ctps:
        raise ValueError("Error in `_prepare_ctps_numba`: `ctps_out` and `ctps` have same number of rows.")
    if n_wanted > n_ctps:
        raise ValueError("Error in `_prepare_ctps_numba`: More CTPs in `ctps_wanted` than in `ctps`.")    
    if not ctps_wanted.shape[1] == n_blocks:
        raise ValueError("Error in `_prepare_ctps_numba`: `ctps` and `ctps_wanted` must have same number of columns.")
    if not ctps_out.shape[1] == n_blocks:
        raise ValueError("Error in `_prepare_ctps_numba`: `ctps` and `ctps_out` must have same number of columns.")

    # allocate array for where cetain ctps_wanted is in ctps
    indices_wanted = np.zeros(n_wanted, dtype=np.uint64)
    indices_wanted.fill(n_ctps)
    
    # loop over all wanted ctps
    for idx_wanted in range(n_wanted):
        # loop over all ctps and check, whether it equals the wanted ctp
        for idx_ctp in range(n_ctps):
            for k in range(n_blocks):
                if ctps[idx_ctp,k] != ctps_wanted[idx_wanted,k]:
                    break
            else:
                # check if not occured already, then copy index
                if indices_wanted[idx_wanted] == n_ctps:
                    indices_wanted[idx_wanted] = idx_ctp
                else:
                    raise ValueError("Error in `_prepare_ctps_numba`: Wanted CTP in "
                                     "`ctps_wanted` appears more than once in `ctps`.") 
        
        # check if this wanted ctp has been found anywhere
        if indices_wanted[idx_wanted] == n_ctps:
            raise ValueError("Error in `_prepare_ctps_numba`: Wanted CTP in "
                             "`ctps_wanted` does not appear in `ctps`.") 
    
    # check if all wanted ctps are unique by checking their indices
    for i in range(n_wanted):
        for j in range(i+1,n_wanted):
            if indices_wanted[i] == indices_wanted[j]:
                raise ValueError("Error in `_prepare_ctps_numba`: Not all wanted "
                                 "CTPs in `ctps_wanted` are unique.") 
    
    
    # copy the wanted ctps to front of output array
    for i in range(n_wanted):
        ctps_out[i,:] = ctps[indices_wanted[i],:]
    
    # sort indices of wanted ctps to make second copy step work
    indices_wanted.sort()
    
    # again loop over all ctps to copy to output array
    idx_out    = n_wanted
    idx_wanted = 0
    for i in range(n_ctps):
        
        # skip if current ctp is one of the wanted ones (already in output)
        if idx_wanted < n_wanted:
            if indices_wanted[idx_wanted] == i:
                idx_wanted += 1
                continue
        
        # copy to the current line
        ctps_out[idx_out,:] = ctps[i,:]
        idx_out += 1



#################################################################################
#                                                                               #
#   SSSS   III    GGGG  N   N   AAA   L              CCCC   AAA   L       CCCC  #
#  S        I    G      NN  N  A   A  L             C      A   A  L      C      #
#   SSS     I    G  GG  N N N  AAAAA  L             C      AAAAA  L      C      #
#      S    I    G   G  N  NN  A   A  L             C      A   A  L      C      #
#  SSSS    III    GGG   N   N  A   A  LLLLL          CCCC  A   A  LLLLL   CCCC  #
#                                                                               #
#################################################################################



def get_S0(ctps, ctps_wanted):
    """
    Construct the reference signal for desired CTPs from a list of CTPs.
    
    S0 is 1 for desired CTPs and 0 elsewise.    
    
    Parameters
    ----------
    ctps : array_like, shape (Nctps, Nblocks)
        Coherence transfer paths.
    ctps_wanted: array_like, shape (Nwanted, Nblocks)
        Desired coherence transfer paths.
    
    Returns
    -------
    S0 : array, shape (Nctp,)
        Array with 1 at wanted CTPs ans 0 elsewise.    
    """
    
    # cast to numpy arrays
    ctps = np.asarray(ctps)
    ctps_wanted = np.asarray(ctps_wanted)
    
    # check shape and dimensions
    if not ctps.ndim == 2:
        raise ValueError('ctps must be 2D!')
    if not ctps_wanted.ndim == 2:
        raise ValueError("ctps_wanted must be 2D!")
    if not ctps.shape[1] == ctps_wanted.shape[1]:
        raise ValueError('ctps and ctps_wanted must have identical shape in last dimension.')
    
    # allocate space
    S0 = np.zeros(ctps.shape[0])
    for idx, ctp_wanted in enumerate(ctps_wanted):
        areequal = (ctps==ctp_wanted).all(axis=1)
        S0[areequal] = 1
        
    return S0



def calculate_signal_rec(phases, rec, dctps):
    """
    Calculate complex signal of coherence transfer paths for given pulse and 
    receiver phases.
    
    Parameters
    ----------
    phases : array_like, shape (Nscans, Nblocks-1)
        Phases for each pulse block an scan.
    rec : array_like, shape (Nscans,)
        Receiver phase for each scan.
    dctps : array_like, shape (Nctps, Nblocks-1)
        Differences in coherence order for each coherence transfer pathway and
        pulse block
    
    Returns
    -------
    signal : array, shape (Nctps,)
        Signal vector averaged over all scans.
    """
    
    return np.exp(-1.0j*( np.matmul(dctps, phases.T) + rec[None,:] )).mean(axis=1)



def calculate_signal_dctps(phases, dctp_ref, dctps):
    """   
    Calculate complex signal of coherence transfer paths for given pulse phases
    and a reference coherence transfer path.
    
    Parameters
    ----------
    phases : array_like, shape (Nscans, Nblocks-1)
        Phases for each pulse block an scan.
    dctp_ref : array_like, shape (Nblocks-1,)
        Reference coherence transfer path.
    dctps : array_like, shape (Nctps, Nblocks-1)
        Differences in coherence order for each coherence transfer pathway and
        pulse block.
    
    Returns
    -------
    signal : array, shape (Nctps,)
        Signal vector averaged over all scans.    
    """
    
    return np.exp(-1.0j*( np.matmul(dctps-dctp_ref[None, :], phases.T))).mean(axis=1)



def calculate_signal_dctps0(phases, dctps0):
    """   
    Calculate complex signal of coherence transfer paths for given pulse phases
    and a reference coherence transfer path.
    
    Parameters
    ----------
    phases : array_like, shape (Nscans, Nblocks-1)
        Phases for each pulse block an scan.
    dctps0 : array_like, shape (Nctps, Nblocks-1)
        Differences in coherence order for each coherence transfer pathway and
        pulse block minus the differences in coherence order for the reference
        path whose phase is held constant by the receiver.
    
    Returns
    -------
    signal : array, shape (Nctps,)
        Signal vector averaged over all scans.    
    """
    
    return np.exp(-1.0j*( np.matmul(dctps0, phases.T))).mean(axis=1)





###################################################################
#                                                                 #
#   CCCC  Y   Y   CCCC  L      EEEEE          GGGG  EEEEE  N   N  #
#  C       Y Y   C      L      E             G      E      NN  N  #
#  C        Y    C      L      EEE           G  GG  EEE    N N N  #
#  C        Y    C      L      E             G   G  E      N  NN  #
#   CCCC    Y     CCCC  LLLLL  EEEEE          GGG   EEEEE  N   N  #
#                                                                 #
###################################################################

def generate_phasecycle_lists(cycles, reccycle):
    """Get phase cycle table from individual phase cycles.
    
    Parameters
    ----------
    cycles : list of list
    reccycle : list

    Returns
    -------
    phases : 2D array
        List of phases for each block. First dimension
        encodes the phases for each scan, second dimenson
        encodes the phase for each block
    receiver : 1D array
        receiver phase to select the given CTP.    
    """

    # check if all cycles have correct shape
    cycles_conv = []
    for cycle in cycles:
        cycle = np.array(cycle, dtype=np.float64)
        if not cycle.ndim == 1:
            raise ValueError("Each phase cycle in the cycle list must be a 1D array or list!")
        cycles_conv.append(cycle)
    cycles = cycles_conv

    # check if receiver has correct format
    reccycle = np.array(reccycle, dtype=np.float64)
    if not reccycle.ndim == 1: 
        raise ValueError("Receiver phase cycle must be a 1D list or array!")

    # number of pulse blocks
    n_blocks = len(cycles)

    # get length of all individual phase cycles and receiver phase
    lengths = [c.shape[0] for c in cycles]
    lengths.append(reccycle.shape[0])

    # get lowest common multiple of all phase cycle lengths
    n_scans = 1
    for length in lengths:
        n_scans = np.lcm(n_scans, length)

    # construct data arrays
    phases = np.zeros(shape=(n_scans, n_blocks), dtype=np.float64)
    for idx, cycle in enumerate(cycles):
        phases[:, idx] = np.tile(cycle, n_scans//lengths[idx])

    receiver = np.tile(reccycle, n_scans//reccycle.shape[0])

    # sanity check if receiver an pulse blocks need same number of scans
    if not receiver.shape[0] == phases.shape[0]:
        raise ValueError("Receiver and pulse blocks do not use the same number of scans and you can do nothing to correct it!")


    return phases, receiver



def generate_phcycle_nested(nblocks, steplengths, pulses, pathway, unit='rad'):
    """
    Generate a nested phase cycle for a pulse sequence with n blocks.
    
    Paramters
    ---------
    nblocks : int
        Number of pulse blocks in the pulse sequence.
    steplengths : list of size 'nblocks'
        Number of steps for the particular step's phase cycle. For 4, the
        corresponding phase cycle would go through 0°, 90°, 180° and 270°.
    pulses : lists of 'nblocks' lists or None
        Indices of pulses that are cycles in a certain step. Start counting
        from 0! If None, each block is cycled one after another.
    pathway : array_like, shape (nblock,)
        Coherence order changes for the desired pathway.
    unit : str, optional
        Unit for the phases. Either 'rad' und 'deg' for radient or degree.
        Defaults to 'rad'.
    
    Returns
    -------
    phases : ndarray, shape (nscans, nblocks)
        List of phases for each block. First dimension
        encodes the phases for each scan, second dimenson
        encodes the phase for each block
    receiver : array, shape (nscans,)
        receiver phase to select the given CTP.
    """
    
    steplengths = np.asarray(steplengths)
    pathway = np.asarray(pathway)
    
    if pathway.size != nblocks:
        raise ValueError("Number of steps in the desired pathway must be identical to number of blocks!")
    
    # total necessary number of scans
    nscans = np.prod(steplengths)
    
    # cumulative products of step lenghts
    cumprod = np.concatenate(([1], np.cumprod(steplengths)))
    cumprod_inv = np.concatenate(([1], np.cumprod(steplengths[::-1])))

    # do some sanity checks
    
    # allocate space
    phases = np.zeros(shape=(nscans, nblocks))
    
    for idx, nstep in enumerate(steplengths):
        
        ph_increment = np.repeat(np.linspace(0.0, 2.0*np.pi, nstep, endpoint=False), cumprod[idx])
        ph_increment = np.tile(ph_increment, cumprod_inv[-2-idx])
        phases[:,pulses[idx]] += ph_increment[:,None]
    
    phases = np.mod(phases, 2.0*np.pi)
    
    # compute receiver phase
    receiver = np.mod(-np.sum(phases*pathway[None,:], axis=1), 2.0*np.pi)
        
    
    # convert to desired unit
    if unit in ['rad', 'radiant']:
        pass
    elif unit in ['deg', 'degree', '°']:
        phases *= 180.0/np.pi
        receiver *= 180.0/np.pi
    else:
        raise ValueError("Unkown phase unit '{}', choose 'degree' or 'radiant'.".format(unit))
    
    
    return phases, receiver



# def generate_phasecycle_nested(lengths, delta_ps):
#     """Generate phases of nested phase cycle.
    
#     Parameters
#     ----------
#     lengths : list
#         Length of the phase cycles for each block.
#     delta_ps : list
#         One of the desired coherence changes for each
#         block. Needed to construct receiver phase.
            
#     Returns
#     -------
#     phases : 2D array
#         List of phases for each block. First dimension
#         encodes the phases for each scan, second dimenson
#         encodes the phase for each block
#     receiver : 1D array
#         receiver phase to select the given CTP.    
#     """
    
#     lengths = np.array(lengths)
#     delta_ps = np.array(delta_ps)
#     total_length = np.prod(lengths)
#     if lengths.size!=delta_ps.size:
#         raise ValueError("Number of phase cycling blocks does not equal number of coherence changes.")
#     steps = lengths.size
    
#     phases = np.zeros(shape=(total_length, steps))
    
#     for idx,length in enumerate(lengths):
        
#         before = np.prod(lengths[:-(steps-idx)])
#         after = np.prod(lengths[idx+1:])
#         k = np.linspace(0,length-1,int(length))
#         phases[:,idx] = 2.0*np.pi/length * np.tile(np.repeat(k, after), before)
    
#     receiver = np.mod(-(phases*delta_ps[None,:]).sum(axis=1), 2.0*np.pi)
    
#     return phases, receiver



def generate_phasecycle_cogwheel(Ns, windings, winding_rec, unit='rad'):
    """Generate phases of a cogwheel phase cycle.
    
    Parameters
    ----------
    Ns : int
        Number of scans for the cogwheel phase cycle.
    windings : list
        List of winding numbers for each pule block.
    winding_rec : int
        Winding number of receiver.
    
    Returns
    -------
    phases : 2D array
        List of phases for each block. First dimension encodes the phases for 
        each scan, second dimenson encodes the phase for each block
    receiver : 1D array
        Receiver phase to select the given CTP.    
    """
    
    windings = np.array(windings)
    steps = np.linspace(0,Ns-1,Ns)
    
    phases = np.mod(2.0*np.pi/Ns * steps[:,None]*windings[None,:], 2.0*np.pi)
    receiver = np.mod(2.0*np.pi*winding_rec*steps/Ns, 2.0*np.pi)

    # convert to desired unit
    if unit in ['rad', 'radiant']:
        pass
    elif unit in ['deg', 'degree', '°']:
        phases *= 180.0/np.pi
        receiver *= 180.0/np.pi
    else:
        raise ValueError("Unkown phase unit '{}', choose 'degree' or 'radiant'.".format(unit))
    
    return phases, receiver








import numpy as np
import numba

############################################################################################
#                                                                                          #
#    GGGG  RRRR    AAA   DDDD    III   EEEEE  N   N  TTTTT          OOO   PPPP   TTTTT     #
#   G      R   R  A   A  D   D    I    E      NN  N    T           O   O  P   P    T       #
#   G  GG  RRRR   AAAAA  D   D    I    EEE    N N N    T           O   O  PPPP     T       #
#   G   G  R  R   A   A  D   D    I    E      N  NN    T           O   O  P        T       #
#    GGG   R   R  A   A  DDDD    III   EEEEE  N   N    T            OOO   P        T       #
#                                                                                          #
############################################################################################

def cost_delta2(S, S0):
    """Cost function for given signal and refernce Signal.
    
    Absolute possible minimum is 0.
    """
    
    return np.mean(np.square(np.abs(S-S0)))




def phfree_cost(phases, dctps0, S0):
    """Cost function for given signal and refernce Signal.
    
    Absolute possible minimum is 0.
    """
    
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size/Nb)
    
    exponent = np.matmul(dctps0, phases.reshape(Ns,Nb).T)
    S = np.exp(-1j*exponent).mean(axis=1)
    
    return np.mean(np.square(np.abs(S-S0)))



def phfree_jac(phases, dctps0, S0):
    """Das ist richtig!..."""
    
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size/Nb)
    
    exponent = np.exp(-1j*np.matmul(dctps0, phases.reshape(Ns,Nb).T)) # has shape (Np, Ns)
    dSstar = np.conj((exponent.mean(axis=1)-S0))                      # has shape (Np,)
    
    # g has (Ns, Nb)
    g = 2.0/(Np*Ns) * np.matmul( np.imag( exponent*dSstar[:,None]).T, dctps0 )
    
    return g.flatten()


def phfree_jacnum(phases, dctps0, S0, dph=1e-5):
    """
    Numerical gradient for the selectivity of a cogwheel phase cylce.
    
    Parameters
    ----------
    windings : (Nb,) array_like
        Winding numbers for each block.
    N : int
        Number of scans.
    dCTPs0 : (Np,Nb) array_like
        Changes of coherence order for each block and path relative to the
        changes of coherence order for the reference path.
    S0 : (Np,) array_like
        Vector of wanted signal for each CTP averaged over all scans. Set the
        elements to 1 for a wanted pathway and 0 for an unwanted.
    
    Returns
    -------
    cost : float
        Value of the cost function.    
    """
    
    phases = np.copy(phases)
    jac = np.zeros(phases.shape)
    
    for i in range(jac.shape[0]):
        for j in range(jac.shape[1]):
            phases[i,j] += dph
            cost_p = phfree_cost(phases, dctps0, S0)
            phases[i,j] -= 2.0*dph
            cost_m = phfree_cost(phases, dctps0, S0)
            phases[i,j] += dph
            jac[i,j] = 0.5*(cost_p-cost_m)/dph
    
    return jac


def phfree_costjac(phases, dctps0, S0):
    """Das ist richtig!..."""
    
    # phases has shape (Ns, Nb)
    # dctps0 has shape (Np, Nb)
    
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size/Nb)
    
    exponent = np.matmul(dctps0, phases.reshape(Ns,Nb).T) # has shape (Np, Ns)
    dS = np.exp(-1j*exponent).mean(axis=1) - S0           # has shape (Np,)
    
    # g has (Ns, Nb)
    g = 2.0/(Np*Ns) * np.matmul( np.imag(np.exp(-1j*exponent)*np.conj(dS)[:,None]).T, dctps0 )
    
    return np.mean(np.square(np.abs(dS))), g.flatten()



def phfree_hessiannum(phases, dctps0, S0, dph=1e-6):
    
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size/Nb)
    
    phases = np.copy(phases).flatten()
    hessian = np.zeros(shape=(phases.size, phases.size))
    
    for i in range(phases.size):
        # diagonal element
        cost_0 = phfree_cost(phases, dctps0, S0)
        phases[i] += dph
        cost_p = phfree_cost(phases, dctps0, S0)
        phases[i] -= 2.0*dph
        cost_m = phfree_cost(phases, dctps0, S0)
        
        hessian[i,i] = (cost_p-2.0*cost_0+cost_m)/dph**2
        
        for j in range(i+1,phases.size):
            # off diagonal elements
            phases[j] -= dph
            cost_mm = phfree_cost(phases, dctps0, S0)
            phases[j] += 2.0*dph
            cost_mp = phfree_cost(phases, dctps0, S0)
            phases[i] += 2.0*dph
            cost_pp = phfree_cost(phases, dctps0, S0)
            phases[j] -= 2.0*dph
            cost_pm = phfree_cost(phases, dctps0, S0)
            
            hessian[i,j] = hessian[j,i] = 0.25*(cost_pp-cost_pm-cost_mp+cost_mm)/dph**2
            phases[i] -= dph
            phases[j] += dph
    
    return hessian











# numba compiled functions for various

@numba.njit(parallel=False)
def phfree_cost_numba(phases, dctps0, S0):
    """Cost function for given signal and refernce Signal.
    
    Absolute possible minimum is 0.
    """
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size//Nb)
    
    S0real = S0.real
    S0imag = S0.imag
    
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    # compute signal and cost
    cost = 0.0
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = 0.0
            for k in range(Nb):
                phshift += phases[j,k] * dctps0[i,k]
            
            S_real += np.cos(phshift)
            S_imag -= np.sin(phshift)
        
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += np.square(dS_real) + np.square(dS_imag) 
        
    return cost/Np



@numba.njit(parallel=False)
def phfree_jac_numba(phases, dctps0, S0):
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size//Nb)
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    S0real = S0.real
    S0imag = S0.imag
    
    # compute jac
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = 0.0
            for k in range(Nb):
                phshift += phases[j,k] * dctps0[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            for k in range(Nb):
                jac[j,k] += helper*dctps0[i,k]
    
    return 2.0/(Np*Ns)*jac.flatten()



@numba.njit(parallel=False)
def phfree_costjac_numba(phases, dctps0, S0):
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size//Nb)
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    S0real = S0.real
    S0imag = S0.imag
        
    # compute cost and jac
    cost = 0.0
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = 0.0
            for k in range(Nb):
                phshift += phases[j,k] * dctps0[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += np.square(dS_real) + np.square(dS_imag) 
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            for k in range(Nb):
                jac[j,k] += helper*dctps0[i,k]
    
    return cost/Np, 2.0/(Np*Ns)*jac.flatten()



@numba.njit(parallel=False)
def phfree_hess_numba(phases, dctps0, S0):
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size/Nb)
    
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    # pre-computation signal and exponent
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    dS_real = np.zeros(Np)
    dS_imag = np.zeros(Np)
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            exponent = 0.0
            for k in range(Nb):
                exponent += phases[j,k] * dctps0[i,k]      
            
            cos[i,j] = np.cos(exponent)
            sin[i,j] = np.sin(exponent)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
        
        dS_real[i] = S_real/Ns - S0[i].real
        dS_imag[i] = S_imag/Ns - S0[i].imag
    
    # allocate space
    hessian = np.zeros((Ns*Nb, Ns*Nb))
    
    # outer loop over all j blocks
    for j1 in numba.prange(Ns):
        
        # diagonal block j1 = j2
        for k1 in range(Nb):
            
            # the diagonal element
            element = 0.0
            for i in range(Np):
                element +=  dctps0[i,k1]*dctps0[i,k1] * \
                    ( 1.0 + Ns*(sin[i,j1]*dS_imag[i]-cos[i,j1]*dS_real[i]) )
            element *= 2.0/(Np*Ns**2)
            hessian[j1*Nb+k1,j1*Nb+k1] = element
            
            # all the other elements
            for k2 in range(k1+1,Nb):
                
                element = 0.0
                for i in range(Np):
                    element += dctps0[i,k1]*dctps0[i,k2] * \
                        ( 1.0 + Ns*(sin[i,j1]*dS_imag[i]-cos[i,j1]*dS_real[i]) )
                element *= 2.0/(Np*Ns**2)
                hessian[j1*Nb+k1,j1*Nb+k2] = element
                hessian[j1*Nb+k2,j1*Nb+k1] = element
                
        
        # the other blocks j1 != j2
        for j2 in range(j1+1,Ns):
            for k1 in range(Nb):
                
                # the diagonal element
                element = 0.0
                for i in range(Np):
                    element +=  dctps0[i,k1]*dctps0[i,k1] * \
                        (cos[i,j1]*cos[i,j2] + sin[i,j1]*sin[i,j2])
                element *= 2.0/(Np*Ns**2)
                hessian[j1*Nb+k1,j2*Nb+k1] = element
                hessian[j2*Nb+k1,j1*Nb+k1] = element
                
                # all the other elements
                for k2 in range(k1+1,Nb):
                    
                    element = 0.0
                    for i in range(Np):
                        element += dctps0[i,k1]*dctps0[i,k2] * \
                            (cos[i,j1]*cos[i,j2] + sin[i,j1]*sin[i,j2])
                    element *= 2.0/(Np*Ns**2)
                    hessian[j1*Nb+k1,j2*Nb+k2] = element
                    hessian[j1*Nb+k2,j2*Nb+k1] = element
                    hessian[j2*Nb+k1,j1*Nb+k2] = element
                    hessian[j2*Nb+k2,j1*Nb+k1] = element                    
    
    return hessian



@numba.njit(parallel=False)
def phfree_hessp_numba(phases, p, dctps0, S0):
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size/Nb)
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    p = p.reshape((Ns,Nb))  
        
    # pre-compute
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    dS_real = np.zeros(Np)
    dS_imag = np.zeros(Np)
    dctps0_p = np.zeros((Np,Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            exponent = 0.0
            for k in range(Nb):
                exponent += phases[j,k] * dctps0[i,k]
                dctps0_p[i,j] += dctps0[i,k] * p[j,k]
            
            cos[i,j] += np.cos(exponent)
            sin[i,j] += np.sin(exponent)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real[i] = S_real/Ns - S0[i].real
        dS_imag[i] = S_imag/Ns - S0[i].imag
    
    # compute Hessian*p
    hessp = np.zeros(phases.shape)
    for j in numba.prange(Ns):
        for k in range(Nb):
            
            for i in range(Np):
                sum_cos = 0.0
                sum_sin = 0.0
                for J in range(Ns):
                    sum_cos += dctps0_p[i,J] * cos[i,J]
                    sum_sin += dctps0_p[i,J] * sin[i,J]

                sum_cos -= Ns * dctps0_p[i,j] * dS_real[i]
                sum_sin += Ns * dctps0_p[i,j] * dS_imag[i]
                hessp[j,k] += dctps0[i,k] * (cos[i,j]*sum_cos + sin[i,j]*sum_sin)
    
    return 2.0/(Np*Ns**2) * hessp.flatten()









@numba.njit(parallel=False)
def phfreerec_cost_numba(phases, dctps, S0):
    """Cost function for given signal and refernce Signal.
    
    Absolute possible minimum is 0.
    """
    
    # get the dimension of all arrays
    Np, Nb = dctps.shape[0], dctps.shape[1]
    Ns = int(phases.size//(Nb+1))
    
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb+1))
    
    S0real = S0.real
    S0imag = S0.imag
    
    # compute signal and cost
    cost = 0.0
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = phases[j,Nb]
            for k in range(Nb):
                phshift += phases[j,k] * dctps[i,k]
            
            S_real += np.cos(phshift)
            S_imag -= np.sin(phshift)
        
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += np.square(dS_real) + np.square(dS_imag) 
        
    return cost/Np



@numba.njit(parallel=False)
def phfreerec_jac_numba(phases, dctps, S0):
    
    # get the dimension of all arrays
    Np, Nb = dctps.shape[0], dctps.shape[1]
    Ns = int(phases.size//(Nb+1))
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb+1))
    
    S0real = S0.real
    S0imag = S0.imag
    
    # compute jac
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = phases[j,Nb]
            for k in range(Nb):
                phshift += phases[j,k] * dctps[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            jac[j,Nb] += helper
            for k in range(Nb):
                jac[j,k] += helper*dctps[i,k]
    
    return 2.0/(Np*Ns)*jac.flatten()



@numba.njit(parallel=False)
def phfreerec_costjac_numba(phases, dctps, S0):
    
    # get the dimension of all arrays
    Np, Nb = dctps.shape[0], dctps.shape[1]
    Ns = int(phases.size//(Nb+1))
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb+1))
    
    S0real = S0.real
    S0imag = S0.imag
        
    # compute cost and jac
    cost = 0.0
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = phases[j,Nb]
            for k in range(Nb):
                phshift += phases[j,k] * dctps[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += np.square(dS_real) + np.square(dS_imag) 
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            jac[j,Nb] += helper
            for k in range(Nb):
                jac[j,k] += helper*dctps[i,k]
    
    return cost/Np, 2.0/(Np*Ns)*jac.flatten()








@numba.njit(parallel=False)
def phfreew_cost_numba(phases, dctps0, S0, weights):
    """Cost function for given signal and refernce Signal.
    
    Absolute possible minimum is 0.
    """
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size//Nb)
    
    S0real = S0.real
    S0imag = S0.imag
    
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    # compute signal and cost
    cost = 0.0
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = 0.0
            for k in range(Nb):
                phshift += phases[j,k] * dctps0[i,k]
            
            S_real += np.cos(phshift)
            S_imag -= np.sin(phshift)
        
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += weights[i]*(np.square(dS_real) + np.square(dS_imag) )
        
    return cost/Np



@numba.njit(parallel=False)
def phfreew_jac_numba(phases, dctps0, S0, weights):
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size//Nb)
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    S0real = S0.real
    S0imag = S0.imag
    
    # compute jac
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = 0.0
            for k in range(Nb):
                phshift += phases[j,k] * dctps0[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -weights[i]*(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            for k in range(Nb):
                jac[j,k] += helper*dctps0[i,k]
    
    return 2.0/(Np*Ns)*jac.flatten()



@numba.njit(parallel=False)
def phfreew_costjac_numba(phases, dctps0, S0, weights):
    
    # get the dimension of all arrays
    Np, Nb = dctps0.shape[0], dctps0.shape[1]
    Ns = int(phases.size//Nb)
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb))
    
    S0real = S0.real
    S0imag = S0.imag
        
    # compute cost and jac
    cost = 0.0
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = 0.0
            for k in range(Nb):
                phshift += phases[j,k] * dctps0[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += weights[i]*(np.square(dS_real) + np.square(dS_imag))
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -weights[i]*(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            for k in range(Nb):
                jac[j,k] += helper*dctps0[i,k]
    
    return cost/Np, 2.0/(Np*Ns)*jac.flatten()










@numba.njit(parallel=False)
def phfreewrec_cost_numba(phases, dctps, S0, weights):
    """Cost function for given signal and refernce Signal.
    
    Absolute possible minimum is 0.
    """
    
    # get the dimension of all arrays
    Np, Nb = dctps.shape[0], dctps.shape[1]
    Ns = int(phases.size//(Nb+1))
    
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb+1))
    
    S0real = S0.real
    S0imag = S0.imag
    
    # compute signal and cost
    cost = 0.0
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = phases[j,Nb]
            for k in range(Nb):
                phshift += phases[j,k] * dctps[i,k]
            
            S_real += np.cos(phshift)
            S_imag -= np.sin(phshift)
        
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += weights[i]*(np.square(dS_real) + np.square(dS_imag))
        
    return cost/Np



@numba.njit(parallel=False)
def phfreewrec_jac_numba(phases, dctps, S0, weights):
    
    # get the dimension of all arrays
    Np, Nb = dctps.shape[0], dctps.shape[1]
    Ns = int(phases.size//(Nb+1))
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb+1))
    
    S0real = S0.real
    S0imag = S0.imag
    
    # compute jac
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = phases[j,Nb]
            for k in range(Nb):
                phshift += phases[j,k] * dctps[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -weights[i]*(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            jac[j,Nb] += helper
            for k in range(Nb):
                jac[j,k] += helper*dctps[i,k]
    
    return 2.0/(Np*Ns)*jac.flatten()


@numba.njit(parallel=False)
def phfreewrec_costjac_numba(phases, dctps, S0, weights):
    
    # get the dimension of all arrays
    Np, Nb = dctps.shape[0], dctps.shape[1]
    Ns = int(phases.size//(Nb+1))
    # reshape phases so it can be used with 1D or 2D array as input
    phases = phases.reshape((Ns,Nb+1))
    
    S0real = S0.real
    S0imag = S0.imag
        
    # compute cost and jac
    cost = 0.0
    jac = np.zeros(phases.shape)
    cos = np.zeros((Np, Ns))
    sin = np.zeros((Np, Ns))
    for i in numba.prange(Np):
        S_real = 0.0
        S_imag = 0.0
        for j in range(Ns):
            
            phshift = phases[j,Nb]
            for k in range(Nb):
                phshift += phases[j,k] * dctps[i,k]      
            
            cos[i,j] += np.cos(phshift)
            sin[i,j] += np.sin(phshift)
            S_real += cos[i,j]
            S_imag -= sin[i,j]
            
        dS_real = S_real/Ns - S0real[i]
        dS_imag = S_imag/Ns - S0imag[i]
        cost += weights[i]*(np.square(dS_real) + np.square(dS_imag))
        
        # compute elements of jacobian
        for j in range(Ns):
            helper = -weights[i]*(cos[i,j]*dS_imag + sin[i,j]*dS_real)
            jac[j,Nb] += helper
            for k in range(Nb):
                jac[j,k] += helper*dctps[i,k]
    
    return cost/Np, 2.0/(Np*Ns)*jac.flatten()



# @numba.njit(parallel=False)
# def phfree_cost_numba(phases, dctps0, S0):
#     """Cost function for given signal and refernce Signal.
    
#     Absolute possible minimum is 0.
#     """
    
#     # get the dimension of all arrays
#     Np, Nb = dctps0.shape[0], dctps0.shape[1]
#     Ns = int(phases.size/Nb)
    
#     # reshape phases so it can be used with 1D or 2D array as input
#     phases = phases.reshape((Ns,Nb))
    
#     # compute signal and cost
#     cost = 0.0
#     for i in numba.prange(Np):
#         S_real = 0.0
#         S_imag = 0.0
#         for j in range(Ns):
            
#             exponent = 0.0
#             for k in range(Nb):
#                 exponent += phases[j,k] * dctps0[i,k]
            
#             S_real += np.cos(exponent)
#             S_imag -= np.sin(exponent)
        
#         dS_real = S_real/Ns - S0[i].real
#         dS_imag = S_imag/Ns - S0[i].imag
#         cost += np.square(dS_real) + np.square(dS_imag) 
        
#     return cost/Np


# @numba.njit(parallel=False)
# def phfree_jac_numba(phases, dctps0, S0):
    
#     # get the dimension of all arrays
#     Np, Nb = dctps0.shape[0], dctps0.shape[1]
#     Ns = int(phases.size/Nb)
#     # reshape phases so it can be used with 1D or 2D array as input
#     phases = phases.reshape((Ns,Nb))  
    
#     # compute jac
#     jac = np.zeros(phases.shape)
#     cos = np.zeros((Np, Ns))
#     sin = np.zeros((Np, Ns))
#     for i in numba.prange(Np):
#         S_real = 0.0
#         S_imag = 0.0
#         for j in range(Ns):
            
#             exponent = 0.0
#             for k in range(Nb):
#                 exponent += phases[j,k] * dctps0[i,k]      
            
#             cos[i,j] += np.cos(exponent)
#             sin[i,j] += np.sin(exponent)
#             S_real += cos[i,j]
#             S_imag -= sin[i,j]
            
#         dS_real = S_real/Ns - S0[i].real
#         dS_imag = S_imag/Ns - S0[i].imag
        
#         # compute elements of jacobian
#         for j in range(Ns):
#             for k in range(Nb):
#                 jac[j,k] -= (cos[i,j]*dS_imag + sin[i,j]*dS_real)*dctps0[i,k]
    
#     return 2.0/(Np*Ns)*jac.flatten()


# @numba.njit(parallel=False)
# def phfree_costjac_numba(phases, dctps0, S0):
    
#     # get the dimension of all arrays
#     Np, Nb = dctps0.shape[0], dctps0.shape[1]
#     Ns = int(phases.size/Nb)
#     # reshape phases so it can be used with 1D or 2D array as input
#     phases = phases.reshape((Ns,Nb))  
        
#     # compute cost and jac
#     cost = 0.0
#     jac = np.zeros(phases.shape)
#     cos = np.zeros((Np, Ns))
#     sin = np.zeros((Np, Ns))
#     for i in numba.prange(Np):
#         S_real = 0.0
#         S_imag = 0.0
#         for j in range(Ns):
            
#             exponent = 0.0
#             for k in range(Nb):
#                 exponent += phases[j,k] * dctps0[i,k]      
            
#             cos[i,j] += np.cos(exponent)
#             sin[i,j] += np.sin(exponent)
#             S_real += cos[i,j]
#             S_imag -= sin[i,j]
            
#         dS_real = S_real/Ns - S0[i].real
#         dS_imag = S_imag/Ns - S0[i].imag
#         cost += np.square(dS_real) + np.square(dS_imag) 
        
#         # compute elements of jacobian
#         for j in range(Ns):
#             for k in range(Nb):
#                 jac[j,k] -= (cos[i,j]*dS_imag + sin[i,j]*dS_real)*dctps0[i,k]
    
#     return cost/Np, 2.0/(Np*Ns)*jac.flatten()



import math
import numpy as np
import numba
from .core.generators import _next_windings_numba
from .core.numbers import _generate_all_factorizations_numba
from .core.array import _copy_by_mask_numba



##########################################################################
#                                                                        #
#   CCCC   OOO   N   N  V   V  EEEEE  N   N   III   EEEEE   CCCC  EEEEE  #
#  C      O   O  NN  N  V   V  E      NN  N    I    E      C      E      #
#  C      O   O  N N N  V   V  EEE    N N N    I    EEE    C      EEE    #
#  C      O   O  N  NN   V V   E      N  NN    I    E      C      E      #
#   CCCC   OOO   N   N    V    EEEEE  N   N   III   EEEEE   CCCC  EEEEE  #
#                                                                        #
##########################################################################



def number_of_nestcogs(sublengths, n_blocks, last_zero, no_inverse):
    """
    Compute the number of nested cogwheel phase cycles.
    
    Parameters
    ----------
    sublengths : (n_subcyc,) array_like
        Number of scans for the cogwheel cycle.
    n_blocks : int
        Number of pulse blocks.
    no_inverse : bool
        Whether to avoid phase inverted counterparts.
    last_zero : bool
        Whether the last winding number shoud be kept zero. This is reasonable
        if all CTPs have the same total coherence order change as the reference
        path because then only the differences between winding numbers do
        really matter.
    
    Returns
    -------
    n_nestcog : int
        Number of nested cogwheel cycles.    
    """
    
    sublengths = np.asarray(sublengths, dtype=int)
    
    if np.any(sublengths < 1):
        raise ValueError("All subcycle lengths must be at least 1.")
    if not np.any(sublengths > 1):
        return 0
    
    if last_zero:
        n_blocks = n_blocks - 1
    
    lengths, counts = np.unique(sublengths, return_counts=True)
    
    n_nestcog = 1
    for i in range(lengths.size):
        
        if lengths[i] == 1:
            continue
        
        if no_inverse:
            n_subcog = (lengths[i]**n_blocks + (2 - lengths[i]%2)**n_blocks)//2 - 1
        else:
            n_subcog = lengths[i]**n_blocks - 1
        
        if n_subcog < counts[i]:
            return 0
        
        acc = n_subcog
        for j in range(1,counts[i]):
            acc = acc*(n_subcog-j) // (j+1)
        n_nestcog *= acc
    
    return n_nestcog





def nestcog_to_phases(windings, lengths, unit='n_scans'):
    """
    Convert a set of nestcog winding numbers to a list of phases.
    
    Parameters
    ----------
    windings : (n_cycles,n_blocks) array_like of int
        Set of winding numbers, one for each cycled block.
    lengths : (n_cycles,) array of int
        Number of scans. Should be at least 2.
    unit : str (optional)
        Unit of the phases. Default is 'n_scans' (full rotation = n_scans).
        Other possible values are 'rad' or 'radiant' for phase in radian 
        (full rotation = 2pi) or 'deg' or 'degree' (full rotation = 360) for
        degree or 'turn' or 'tr' for turns (full rotation = 1).
    
    Returns
    -------
    phases : (n_blocks, n_scans) array
        Phases for every scan and block in specified unit.
    """
    
    # convert windings and lengths and check dimensionality
    windings = np.asarray(windings)
    lengths = np.asarray(lengths)
    if not windings.ndim == 2:
        raise ValueError("`windings` must be a 2D array.")
    if not lengths.ndim == 1:
        raise ValueError("`lengths` must be a 1D array.")
    n_subcyc = windings.shape[0]
    n_blocks = windings.shape[1]
    if not lengths.size == n_subcyc:
        raise ValueError("Size of `lengths` must match first dimension of `windings`.")
        
    if np.any(lengths<1):
        raise ValueError("All subcycle lengths `lengths` should be at least 1.")
    n_scans = np.prod(lengths)
        
    # allocate array
    phases = np.zeros(shape=(n_scans,n_blocks), dtype=windings.dtype)
    
    # set phases using numba routine
    _nestcog_to_phases_numba(windings, lengths, phases)
    
    # convert to specified units
    if unit.lower() in ['n_scans', 'nscnas', 'scans']:
        return phases
    elif unit.lower() in ['rad', 'radian', 'si']:
        return phases * (2.0*np.pi/n_scans)
    elif unit.lower() in ['deg', 'degree', '°']:
        return phases * (360.0/n_scans)
    elif unit.lower() in ['turn', 'tr', 'pla']:
        return phases * (1.0/n_scans)
    else:
        raise ValueError("Unkown unit '{}', try 'n_scans', 'radian' or 'degree'.".format(unit))





def nestcogs_to_phases(windings, lengths, unit='n_scans'):
    """
    Convert a set of nestcog winding numbers to a list of phases.
    
    Parameters
    ----------
    windings : (n_cycles,n_blocks) array_like of int
        Set of winding numbers, one for each cycled block.
    lengths : (n_cycles,) array of int
        Number of scans. Should be at least 2.
    unit : str (optional)
        Unit of the phases. Default is 'n_scans' (full rotation = n_scans).
        Other possible values are 'rad' or 'radiant' for phase in radian 
        (full rotation = 2pi) or 'deg' or 'degree' (full rotation = 360) for
        degree or 'turn' or 'tr' for turns (full rotation = 1).
    
    Returns
    -------
    phases : (n_blocks, n_scans) array
        Phases for every scan and block in specified unit.
    """
    
    # convert windings and lengths and check dimensionality
    windings = np.asarray(windings)
    lengths = np.asarray(lengths)
    if not windings.ndim == 3:
        raise ValueError("`windings` must be a 3D array.")
    if not lengths.ndim == 2:
        raise ValueError("`lengths` must be a 2D array.")
    n_cycles = windings.shape[0]
    n_subcyc = windings.shape[1]
    n_blocks = windings.shape[2]
    if not lengths.shape[1] == n_subcyc:
        raise ValueError("Second dimension of `lengths` must match second dimension of `windings`.")
        
    if np.any(lengths<1):
        raise ValueError("All subcycle lengths `lengths` should be at least 1.")
    n_scans = np.prod(lengths[0])
    for idx_cyc in range(n_cycles):
        if not n_scans == np.prod(lengths[idx_cyc]):
            raise ValueError("All cycles must have the same number of scans.")
        
    # allocate array
    phases = np.zeros(shape=(n_cycles,n_scans,n_blocks), dtype=windings.dtype)
    
    # set phases using numba routine
    _nestcogs_to_phases_numba(windings, lengths, phases)
    
    # convert to specified units
    if unit.lower() in ['n_scans', 'nscnas', 'scans']:
        return phases
    elif unit.lower() in ['rad', 'radian', 'si']:
        return phases * (2.0*np.pi/n_scans)
    elif unit.lower() in ['deg', 'degree', '°']:
        return phases * (360.0/n_scans)
    elif unit.lower() in ['turn', 'tr', 'pla']:
        return phases * (1.0/n_scans)
    else:
        raise ValueError("Unkown unit '{}', try 'n_scans', 'radian' or 'degree'.".format(unit))




#####################################################################
#                                                                   #
#   U   U  TTTTT   III   L       III   TTTTT   III   EEEEE   SSSS   #
#   U   U    T      I    L        I      T      I    E      S       #
#   U   U    T      I    L        I      T      I    EEE     SSS    #
#   U   U    T      I    L        I      T      I    E          S   #
#    UUU     T     III   LLLLL   III     T     III   EEEEE  SSSS    #
#                                                                   #
#####################################################################



@numba.njit(['u1(u1[:])', 'u2(u2[:])', 'u4(u4[:])', 'u8(u8[:])',
             'i1(i1[:])', 'i2(i2[:])', 'i4(i4[:])', 'i8(i8[:])'],
            cache=True)
def _smallest_common_divisor_numba(a):
    """
    Find the smallest common divisor greater than one among a list of integers.
    Numbers smaller than one are ignored.
    
    The smallest number greater than one is found in `a`. Then, all possible
    prime factors (using 6k+-1 filter) are searched in ascending order.
    If all numbers (greater than zero) in `a` share this factor, it is the
    smallest common factor.
    
    Parameters
    ----------
    a : array_like of int
        Numbers to get the smallest common divisor of.
    
    Returns
    -------
    smallest_common : int
        Smallest common divisor greater than one, or one not all numbers share
        a factor (or if all numbers are smaller than or equal to one).
    """
    
    if not a.ndim == 1:
        raise ValueError("`a` must be a 1D array.")
    
    n = a.size
    
    # get minimum value greater than one
    a_min = a[0]
    for i in range(n):
        if a[i] > 1:
            a_min = a[i]
            break
    else:
        return 1
    for j in range(i+1,n):
        if a[j] > 1 and a[j] < a_min:
            a_min = a[j]
    
    # divisibility by 2
    if a_min%2 == 0:
        a_min = a_min//2
        # iterate over array elements greater than 0, see if all are divisible
        for i in range(n):
            if a[i] < 1: continue
            if not a[i]%2 == 0: break
        else:
            return 2
        # divide out this factor
        while a_min%2 == 0:
            a_min = a_min//2
    
    # divisibility by 3
    if a_min%3 == 0:
        a_min = a_min//3
        # iterate over array elements greater than 0, see if all are divisible
        for i in range(n):
            if a[i] < 1: continue
            if not a[i]%3 == 0: break
        else:
            return 3
        # divide out this factor
        while a_min%3 == 0:
            a_min = a_min//3
    
    # divisibility by all other prime factors (6k+-1 filter)
    div = 5
    while a_min > 1:
        
        if a_min%div == 0:
            a_min = a_min//div
            # iterate over array elements greater than 0, see if all are divisible
            for i in range(n):
                if a[i] < 1: continue
                if not a[i]%div == 0: break
            else:
                return div
            # divide out this factor
            while a_min%div == 0:
                a_min = a_min//div
            
        if a_min%(div+2) == 0:
            a_min = a_min//(div+2)
            # iterate over array elements greater than 0, see if all are divisible
            for i in range(n):
                if a[i] < 1: continue
                if not a[i]%(div+2) == 0: break
            else:
                return (div+2)
            # divide out this factor
            while a_min%(div+2) == 0:
                a_min = a_min//(div+2)
        
        div += 6
    
    return 1



# @numba.njit(['u1(u1[:])', 'u2(u2[:])', 'u4(u4[:])', 'u8(u8[:])',
#              'i1(i1[:])', 'i2(i2[:])', 'i4(i4[:])', 'i8(i8[:])'],
#             cache=True)
# def _OLD_smallest_common_divisor_numba(numbers):
#     """
#     Find the smallest common divisor greater than one among a list of integers.
    
#     First the greatest common divisor among all numbers is computed, whose
#     divisors are also divisors of all numbers. Then the smallest prime divisor
#     of the gcd is returned, which also is the smallest common divisor greater
#     than one. One is returned if at least two numbers are coprime and zero if
#     all numbers are zero themselves.
    
#     Parameters
#     ----------
#     numbers : array_like of int
#         Numbers to get the smallest common divisor of.
    
#     Returns
#     -------
#     smallest_common : int
#         Smallest common divisor greater than one, one if not all numbers share
#         a factor or zero if all numbers are zero.
#     """
    
#     numbers = numbers.flatten()
#     DTYPE = type(numbers[0])
    
#     # find common gcd of all numbers
#     gcd = DTYPE(0)
#     for i in range(numbers.size):
#         gcd = np.gcd(gcd, numbers[i])
#         if gcd == 1:
#             return 1
        
#     # handle smaller cases
#     if gcd <= 3:
#         if gcd == 0:
#             # if all numbers are zero
#             return 0
#         elif gcd == 1:
#             # if at least one number is one or at least two numbers are coprime
#             return 1
#         elif gcd == 2:
#             return 2
#         elif gcd == 3:
#             return 3
#         else:
#             raise ValueError('Internal error: gcd has invalid value.')
    
#     # smallest prime factor of gcd, check divisibility by 2 and 3
#     if gcd % 2 == 0:
#         return 2
#     if gcd % 3 == 0:
#         return 3
    
#     # check all possible divisors of form 6*n +- 1 up to sqrt(gcd)
#     stop = int(np.floor(np.sqrt(gcd)))
#     i = 7
#     while i <= stop:
#         if gcd % (i-2) == 0:
#             return i-2
#         if gcd % i == 0:
#             return i
#         i += 6
     
#     # this factor might have been skipped by previous loop
#     if i-2 <= stop:
#         if gcd % (i-2) == 0:
#             return i-2
    
#     # if we are here, gcd is prime, so just return it as smallest common factor
#     return gcd


@numba.njit()
def _nestcog_to_phases_numba(windings, lengths, phases):
    """Utility"""
    
    # check dimensionality and shape of input
    if not windings.ndim == 2:
        raise ValueError("Error in `_nestcog_to_phases_numba`: `windings` must be 2D.")
    if not lengths.ndim == 1:
        raise ValueError("Error in `_nestcog_to_phases_numba`: `lengths` must be 1D.")
    n_subcyc = windings.shape[0]
    n_blocks = windings.shape[1]
    if not lengths.size == n_subcyc:
        raise ValueError("Size of `lengths` must match first dimension of `windings`.")
    
    # check subcycle lengths and compute number of scans
    if np.any(lengths<1):
        raise ValueError("Error in `_nestcog_to_phases_numba`: All subcycle lengths "
                         "`lengths` should be at least 1.")
    n_scans = np.prod(lengths)
    
    # check dimensionality and shape of output array
    if not phases.ndim == 2:
        raise ValueError("Error in `_nestcog_to_phases_numba`: `phases` must be 2D.")
    if (not phases.shape[0] == n_scans) or (not phases.shape[1] == n_blocks):
        raise ValueError("Error in `_nestcog_to_phases_numba`: `phases` must be of "
                         "shape (n_scans,n_blocks).")
    
    cumprod = 1
    for idx_sub in range(n_subcyc):
        
        for i in range(1, lengths[idx_sub]):
            increment = n_scans // lengths[idx_sub]
            
            for j in range(cumprod):
                pivot = i*cumprod+j
            
                for idx_block in range(n_blocks):
                    phases[pivot,idx_block] = (phases[pivot-cumprod,idx_block] 
                                               + windings[idx_sub, idx_block]*increment)%n_scans
        
        cumprod *= lengths[idx_sub]



@numba.njit(parallel=True)
def _nestcogs_to_phases_numba(windings, lengths, phases):
    """Utility"""
    
    # check dimensionality and shape of input
    if not windings.ndim == 3:
        raise ValueError("Error in `_nestcogs_to_phases_numba`: `windings` must be 3D.")
    if not lengths.ndim == 2:
        raise ValueError("Error in `_nestcogs_to_phases_numba`: `lengths` must be 2D.")
    n_cycles = windings.shape[0]
    n_subcyc = windings.shape[1]
    n_blocks = windings.shape[2]
    if not lengths.shape[1] == n_subcyc:
        raise ValueError("Error in `_nestcogs_to_phases_numba`: Second dimension "
                         "`lengths` must match second dimension of `windings`.")
    
    # check subcycle lengths and compute number of scans
    if np.any(lengths<1):
        raise ValueError("Error in `_nestcogs_to_phases_numba`: All subcycle lengths "
                          "`lengths` should be at least 1.")
    n_scans = np.prod(lengths[0,:])
    for idx_cyc in range(n_cycles):
        if not n_scans == np.prod(lengths[idx_cyc,:]):
            raise ValueError("Error in `_nestcog_to_phases_numba`: All cycles must have "
                              "the same number of scans.")
    
    # check dimensionality and shape of output array
    if not phases.ndim == 3:
        raise ValueError("Error in `_nestcog_to_phases_numba`: `phases` must be 3D.")
    if (not phases.shape[1] == n_scans) or (not phases.shape[2] == n_blocks):
        raise ValueError("Error in `_nestcog_to_phases_numba`: `phases` must be of "
                          "shape (n_scans,n_blocks).")
        
    for idx_cyc in numba.prange(n_cycles):
        
        _nestcog_to_phases_numba(
            windings[idx_cyc,:,:],
            lengths[idx_cyc,:],
            phases[idx_cyc,:,:])




############################################################
#                                                          #
#   CCCC   OOO   U   U  N   N  TTTTT   III   N   N   GGGG  #
#  C      O   O  U   U  NN  N    T      I    NN  N  G      #
#  C      O   O  U   U  N N N    T      I    N N N  G  GG  #
#  C      O   O  U   U  N  NN    T      I    N  NN  G   G  #
#   CCCC   OOO    UUU   N   N    T     III   N   N   GGG   #
#                                                          #
############################################################



# @numba.njit(['i8(i8[:,:],i8[:],i8[:,:],b1,b1,b1,b1)',
#              'i8(u8[:,:],u8[:],u8[:,:],b1,b1,b1,b1)',
#              'i4(i4[:,:],i4[:],i4[:,:],b1,b1,b1,b1)',
#              'i4(u4[:,:],u4[:],u4[:,:],b1,b1,b1,b1)'
#              'i2(i2[:,:],i2[:],i2[:,:],b1,b1,b1,b1)',
#              'i2(u2[:,:],u2[:],u2[:,:],b1,b1,b1,b1)'],
#             fastmath=True, cache=False)
@numba.njit(fastmath=True)
def _init_nestwind_numba(windings, lengths, factorizations,
            last_zero, no_inverse, no_cfscan, no_cfwind):
    """
    Generate the first set of nestcog parameters, than includes the next
    set of winding numbers and subcycle lengths.
    
    Parameters
    ----------
    windings : (n_factors,n_blocks) 2D array of int
        Winding numbers. The last rows corresponding the lengths of one are
        ignored and should be zero.
    lengths : (n_factors,) 1D array of int
        Subcycle lengths for each subcycle. The lengths must be sorted but the
        last elements can be one and are ignored.
    factorizations : (n_factorizations,n_factors) 2D array of int
        All possible factorizations to consider. If the set of winding numbers
        cannot be counted up, the next factorization is set for lengths.
    last_zero : bool
        Whether the last winding number shoud be kept zero. This is reasonable
        if all CTPs have the same total coherence order change as the reference
        path because then only the differences between winding numbers do
        really matter.
    no_inverse : bool
        Whether the last non-zero winding number is counted up only to
        n_scans//2 and not n_scans-1. This avoids generating the phase inverted
        counterparts of other sets of winding numbers.
    no_cfscan : bool
        Whether to avoid sets of winding numbers, where every winding number and
        the number of scans share a common factor. This would result in an
        redundant cycle, i.e. a cycle that has smaller length but is repeated.
    no_cfwind : bool
        Whether to avoid sets of winding numbers, where the winding numbers 
        share a common factor. Such cycles would be equivalent (selecting the)
        same CTPs as the cycle with winding numbers, where the common factor
        is devided out.
        If this option is set, setting `no_cfscan` has no effect anymore.
    
    Returns
    -------
    status : int
        0 if the set of winding numbers could be initialized. Negative numbers 
        indicate errors.
    """
    
    # check shape and size of input
    if not windings.ndim == 2:
        raise ValueError("Error in `_init_nestwind_numba`: "
                "`windings` must be a 2D array.")
    if not lengths.ndim == 1:
        raise ValueError("Error in `_init_nestwind_numba`: "
                "`lengths` must be a 2D array.")
    if not factorizations.ndim == 2:
        raise ValueError("Error in `_init_nestwind_numba`: "
                "`factorizations` must be a 2D array.")
    if not windings.shape[0] == lengths.size:
        raise ValueError("Error in `_init_nestwind_numba`: "
                "First dimension of `windings` must equal size of `lengths`.")
    if not factorizations.shape[1] == lengths.size:
        raise ValueError("Error in `_init_nestwind_numba`: "
                "Second dimension of `factorizations` must equal size of `lengths`.")
        
    # set some shape parameters
    n_factors        = windings.shape[0]        # maximum number of factors
    n_blocks         = windings.shape[1]        # number of blocks in pulse sequence
    n_factorizations = factorizations.shape[0]  # total number of factorizations 
    
    # try to init windings
    for idx_factorization in range(n_factorizations):
        
        # set the current factorization
        lengths[:] = factorizations[idx_factorization,:]
    
        # try to init for this factorization
        
        # find number of subcycles and check lengths array
        for idx in range(n_factors):
            # check if value 
            if lengths[idx] <= 1:
                # check if value is greater than one
                if lengths[idx] < 1:
                    raise ValueError("Error in `_init_nestwind_numba`: "
                            "All values in `lengths` must be at least 1.")
                # if one, all next numbers must be one as well
                if lengths[idx] == 1:
                    n_subcyc = idx
                    for jdx in range(idx+1, lengths.size):
                        if not lengths[jdx] == 1:
                            raise ValueError("Error in `_init_nestwind_numba`: "
                                    "The last values of `lengths` must all be one.")
                    break
            # check order
            if idx > 0:
                if lengths[idx] < lengths[idx-1]:
                    raise ValueError("Error in `_init_nestwind_numba`: "
                            "`lengths` must be sorted except for the last values equal to 1.")        
        else:
            n_subcyc = n_factors
        
        # init first row
        windings[0,:] = 0
        windings[0,0] = 1
        
        # loop over all other rows and try to init them
        no_overflow = True
        for idx in range(1, n_subcyc):

            # if same cycle lengths, avoid repetition
            if lengths[idx] == lengths[idx-1]:
                # copy previous line and try to count up
                windings[idx,:] = windings[idx-1,:]
                res = _next_windings_numba(windings[idx,:], lengths[idx],
                        last_zero, no_inverse, no_cfscan, no_cfwind)
                if res == 0:
                    continue
                elif res == 1:
                    # overflow happened, try next factorization
                    no_overflow = False
                    break
                else:
                    raise ValueError("Error in `_init_nestwind_numba`: "
                            "Error when counting up winding numbers.")
            
            # if longer cycle, init next row
            elif lengths[idx] > lengths[idx-1]:
                windings[idx,:] = 0
                windings[idx,0] = 1
            
            # cycle lenghts must be ordered
            else:
                raise ValueError("Error in `_init_nestwind_numba`: "
                        "`lengths` must be sorted except for the last values equal to 1.")
        
        if no_overflow:
            windings[n_subcyc:,:] = 0
            return 0
    
    # the loop must break
    else:
        raise ValueError("Error in `_init_nestwind_numba`: "
                "Could not initialize with any factorization.")





# @numba.njit(['i8(i8[:,:],i8[:],i8[:,:],b1,b1,b1,b1)',
#              'i8(u8[:,:],u8[:],u8[:,:],b1,b1,b1,b1)',
#              'i4(i4[:,:],i4[:],i4[:,:],b1,b1,b1,b1)',
#              'i4(u4[:,:],u4[:],u4[:,:],b1,b1,b1,b1)'
#              'i2(i2[:,:],i2[:],i2[:,:],b1,b1,b1,b1)',
#              'i2(u2[:,:],u2[:],u2[:,:],b1,b1,b1,b1)'],
#             fastmath=True, cache=False)
@numba.njit(fastmath=True)
def _next_nestwind_numba(windings, lengths, factorizations,
                    last_zero, no_inverse, no_cfscan, no_cfwind):
    """
    Generate the next set of nestcog parameters, than includes the next
    set of winding numbers and subcycle lengths.
    
    Parameters
    ----------
    windings : (n_factors,n_blocks) 2D array of int
        Winding numbers. The last rows corresponding the lengths of one are
        ignored and should be zero.
    lengths : (n_factors,) 1D array of int
        Subcycle lengths for each subcycle. The lengths must be sorted but the
        last elements can be one and are ignored.
    factorizations : (n_factorizations,n_factors) 2D array of int
        All possible factorizations to consider. If the set of winding numbers
        cannot be counted up, the next factorization is set for lengths.
    last_zero : bool
        Whether the last winding number shoud be kept zero. This is reasonable
        if all CTPs have the same total coherence order change as the reference
        path because then only the differences between winding numbers do
        really matter.
    no_inverse : bool
        Whether the last non-zero winding number is counted up only to
        n_scans//2 and not n_scans-1. This avoids generating the phase inverted
        counterparts of other sets of winding numbers.
    no_cfscan : bool
        Whether to avoid sets of winding numbers, where every winding number and
        the number of scans share a common factor. This would result in an
        redundant cycle, i.e. a cycle that has smaller length but is repeated.
    no_cfwind : bool
        Whether to avoid sets of winding numbers, where the winding numbers 
        share a common factor. Such cycles would be equivalent (selecting the)
        same CTPs as the cycle with winding numbers, where the common factor
        is devided out.
        If this option is set, setting `no_cfscan` has no effect anymore.
    
    Returns
    -------
    status : int
        0 if the set of winding numbers could be counted up, 1 otherwise. In
        that case the last possible set of winding numbers was generated and
        the array should be back to initial configuration. Negative numbers 
        indicate errors.
    """
    
    # check shape and size of input
    if not windings.ndim == 2:
        raise ValueError("Error in `_next_nestwind_numba`: "
                "`windings` must be a 2D array.")
    if not lengths.ndim == 1:
        raise ValueError("Error in `_next_nestwind_numba`: "
                "`lengths` must be a 1D array.")
    if not factorizations.ndim == 2:
        raise ValueError("Error in `_next_nestwind_numba`: "
                "`factorizations` must be a 2D array.")
    if not windings.shape[0] == lengths.size:
        raise ValueError("Error in `_next_nestwind_numba`: "
                "First dimension of `windings` must equal size of `lengths`.")
    if not factorizations.shape[1] == lengths.size:
        raise ValueError("Error in `_next_nestwind_numba`: "
                "Second dimension of `factorizations` must equal size of `lengths`.")
        
    # set some shape parameters
    n_factors        = windings.shape[0]        # maximum number of factors
    n_blocks         = windings.shape[1]        # number of blocks in pulse sequence
    n_factorizations = factorizations.shape[0]  # total number of factorizations 
    
    
    # try to count up count up with current factorization as lengths
    
    # find number of subcycles and check lengths array
    # n_subcyc is the number of factors (sublengths) unequal 1
    for idx in range(lengths.size):
        # if value is not bigger than one
        if lengths[idx] <= 1:
            # value must be greater than or equal to one
            if lengths[idx] < 1:
                return -1
            # if one, all next numbers must be one as well
            if lengths[idx] == 1:
                n_subcyc = idx
                for jdx in range(idx+1, lengths.size):
                    if not lengths[jdx] == 1:
                        return -1
                break
        # check if lengths is sorted
        if idx > 0:
            if lengths[idx] < lengths[idx-1]:
                return -1
    # if loop ran through all values in lengths are greater than one
    else:
        n_subcyc = lengths.size
    
    
    # loop over all cols in reversed order
    for idx in range(n_subcyc-1,-1,-1):
        
        # try to count up the current row
        res = _next_windings_numba(windings[idx,:], lengths[idx],
                    last_zero, no_inverse, no_cfscan, no_cfwind)
        
        # count up successfull, now try to init following rows
        if res == 0:
            
            # loop over successive rows
            for jdx in range(idx+1,n_subcyc):    
                
                # if row before has same subcycle length copy it and count up
                if lengths[jdx] == lengths[jdx-1]:
                    windings[jdx,:] = windings[jdx-1,:]
                    res = _next_windings_numba(windings[jdx,:], lengths[jdx],
                                    last_zero, no_inverse, no_cfscan, no_cfwind)
                    if res == 0:
                        continue
                    elif res == 1:
                        break
                    else:
                        return -1
                
                # row before has smaller subcycle length, make initial configuration
                elif lengths[jdx] > lengths[jdx-1]:
                    windings[jdx,:] = 0
                    windings[jdx,0] = 1
                
                # this should not happen
                else:
                    return -1
            
            # row was counted up and all consecutive ones initialized, work is done
            else:
                return 0
        
        # could not count up this row, so try next one
        elif res == 1:
            continue
        
        else:
            return -1
        
    
    # if could not count up with current factorization, try next one
    
    # get index of recent factorization
    for idx_fac_start in range(n_factorizations):
        # compare all cols
        for jdx in range(n_factors):
            if not factorizations[idx_fac_start,jdx] == lengths[jdx]:
                # try next factorization
                break
        else:
            # if loop ran through, we found the correct line, so break loop
            break
    else:
        # loop must break if current factorization is inside factorizations array
        return -1
    
    # loop over all factorizations that come after the recent one
    for idx_fac in range(idx_fac_start+1, n_factorizations):
        
        # set length to current factorization
        lengths[:] = factorizations[idx_fac,:]
        
        # try to init with this factorization
        
        # find number of subcycles and check lengths array
        # n_subcyc is the number of factors (sublengths) unequal 1
        for idx in range(lengths.size):
            # if value is not bigger than one
            if lengths[idx] <= 1:
                # value must be greater than or equal to one
                if lengths[idx] < 1:
                    return -1
                # if one, all next numbers must be one as well
                if lengths[idx] == 1:
                    n_subcyc = idx
                    for jdx in range(idx+1, lengths.size):
                        if not lengths[jdx] == 1:
                            return -1
                    break
            # check if lengths is sorted
            if idx > 0:
                if lengths[idx] < lengths[idx-1]:
                    return -1
        # if loop ran through all values in lengths are greater than one
        else:
            n_subcyc = lengths.size
            
        # init first row
        windings[0,:] = 0
        windings[0,0] = 1
        
        # loop over all other rows and try to init them
        no_overflow = True
        for idx in range(1, n_subcyc):

            # if same cycle lengths, avoid repetition
            if lengths[idx] == lengths[idx-1]:
                # copy previous line and try to count up
                windings[idx,:] = windings[idx-1,:]
                res = _next_windings_numba(windings[idx,:], lengths[idx],
                        last_zero, no_inverse, no_cfscan, no_cfwind)
                if res == 0:
                    continue
                elif res == 1:
                    # overflow happened, try next factorization
                    no_overflow = False
                    break
                else:
                    return -1
            
            # if longer cycle, init next row
            elif lengths[idx] > lengths[idx-1]:
                windings[idx,:] = 0
                windings[idx,0] = 1
            
            # cycle lenghts must be ordered
            else:
                return -1
        
        if no_overflow:
            windings[n_subcyc:,:] = 0
            return 0
            
    
    # if loop run through last factorization was tried, so we are done
    else:
        res = _init_nestwind_numba(windings, lengths, factorizations,
                    last_zero, no_inverse, no_cfscan, no_cfwind)
        if not res == 0:
            return -1
        else:
            return 1





@numba.njit(fastmath=True)
def _fill_buffer_nestcog_numba(buffer_windings, buffer_lengths,
                               seed_windings, seed_lengths, factorizations,
                               last_zero, no_inverse, no_cfscan, no_cfwind):
    """
    ToDo
    
    Parameters
    ----------
    buffer_windings
    buffer_lengths
    seed_windings
    seed_lengths
    factorizations
    last_zero : bool
        Whether the last winding number shoud be kept zero. This is reasonable
        if all CTPs have the same total coherence order change as the reference
        path because then only the differences between winding numbers do
        really matter.
    no_inverse : bool
        Whether the last non-zero winding number is counted up only to
        n_scans//2 and not n_scans-1. This avoids generating the phase inverted
        counterparts of other sets of winding numbers.
    no_cfscan : bool
        Whether to avoid sets of winding numbers, where every winding number and
        the number of scans share a common factor. This would result in an
        redundant cycle, i.e. a cycle that has smaller length but is repeated.
    no_cfwind : bool
        Whether to avoid sets of winding numbers, where the winding numbers 
        share a common factor. Such cycles would be equivalent (selecting the)
        same CTPs as the cycle with winding numbers, where the common factor
        is devided out.
        If this option is set, setting `no_cfscan` has no effect anymore.
    
    Returns
    -------
    n_filled : int
        Number of filled buffer.
    """
    
    # check shape of input
    if not buffer_windings.ndim == 3:
        raise ValueError("`buffer_windings` must be 3D!")
    if not buffer_lengths.ndim == 2:
        raise ValueError("`buffer_windings` must be 2D!")
    if not seed_windings.ndim == 2:
        raise ValueError("`buffer_windings` must be 2D!")
    if not seed_lengths.ndim == 1:
        raise ValueError("`buffer_windings` must be 1D!")
        
    n_buffered = buffer_windings.shape[0]
    n_factors  = buffer_windings.shape[1]
    n_blocks   = buffer_windings.shape[2]
        
    if not buffer_lengths.shape[0] == n_buffered:
        raise ValueError("1")
    if not buffer_lengths.shape[1] == n_factors:
        raise ValueError("2")
    if not seed_windings.shape[0] == n_factors:
        raise ValueError("3")
    if not seed_windings.shape[1] == n_blocks:
        raise ValueError("4")
    if not seed_lengths.size == n_factors:
        raise ValueError("5")
    if not factorizations.shape[1] == n_factors:
        raise ValueError("6")
        
    # copy seeds to private working arrays
    working_windings = np.zeros_like(seed_windings)
    working_lengths = np.zeros_like(seed_lengths)
    new_seed_windings = np.zeros_like(seed_windings)
    new_seed_lengths = np.zeros_like(seed_lengths)
    working_windings[:,:] = seed_windings[:,:]
    working_lengths[:] = seed_lengths[:]
    
    
    # initialize the first entry
    buffer_windings[0,:,:] = working_windings[:,:]
    buffer_lengths[0,:] = working_lengths[:]
    n_filled = 1
    
    for i in range(1,n_buffered):
        
        # try to count up
        res = _next_nestwind_numba(
                    working_windings, working_lengths, factorizations,
                    last_zero, no_inverse, no_cfscan, no_cfwind)
        if res == 0:
            buffer_windings[i,:,:] = working_windings[:,:]
            buffer_lengths[i,:] = working_lengths[:]
            n_filled += 1
        elif res == 1:
            new_seed_windings[:,:] = working_windings[:,:]
            new_seed_lengths[:] = working_lengths[:]
            return n_filled, new_seed_windings, new_seed_lengths, True
        else:
            return -1, new_seed_windings, new_seed_lengths, False
        
    # check if there is more work to do
    res = _next_nestwind_numba(
                working_windings, working_lengths, factorizations,
                last_zero, no_inverse, no_cfscan, no_cfwind)
    if res == 0:
        all_done = False
    elif res == 1:
        all_done = True
    else:
        return -1, new_seed_windings, new_seed_lengths, False
    
    new_seed_windings[:,:] = working_windings[:,:]
    new_seed_lengths[:] = working_lengths[:]
    return n_filled, new_seed_windings, new_seed_lengths, all_done







#######################################################
#                                                     #
#   V   V   AAA   RRRR    III    OOO   U   U   SSSS   #
#   V   V  A   A  R   R    I    O   O  U   U  S       #
#   V   V  AAAAA  RRRR     I    O   O  U   U   SSS    #
#    V V   A   A  R  R     I    O   O  U   U      S   #
#     V    A   A  R   R   III    OOO    UUU   SSSS    #
#                                                     #
#######################################################



@numba.njit()
def are_ctps_passed_nestcog_numba(cycle_lengths, winding_numbers, dctps0):
    """
    For a given nestcog phase cycle determine if it passes a CTP or blocks it
    given its relative coherence order changes.
    
    Parameters
    ----------
    cycle_lengths : (n_cycles,) array of int
        Cycle lengths of each inner cycle. The number of scans is the product
        all numbers in this array.
    winding_numbers : (n_cycles, n_blocks) array of int
        Winding numbers for each cycle.
    dctps0 : (n_ctps, n_blocks) array of int
        Coherence order changes relative to the coherence order changes of
        the reference path. The desired paths must be given first.
    
    Returns
    -------
    is_passed : array of bool
        For each CTP, wether it is passed by the given phase cycle.
    """
    
    # check size and shape of input
    if not cycle_lengths.ndim == 1:
        raise ValueError('Internal error: `cycle_lengths` must be 1D.')
    if not winding_numbers.ndim == 2:
        raise ValueError('Internal error: `winding_numbers` must be 2D.')
    if not dctps0.ndim == 2:
        raise ValueError('Internal error: `dctps0` must be 2D.')
    if not cycle_lengths.shape[0] == winding_numbers.shape[0]:
        raise ValueError('Internal error: First dimensions of `cycle_lengths` '
                         'and `winding_numbers` must have same size.')
    if not winding_numbers.shape[1] == dctps0.shape[1]:
        raise ValueError('Internal error: Second dimensions of `winding_numbers` '
                         'and `dctps0` must have same size.')
        
    n_cycles = cycle_lengths.shape[0]
    n_ctps = dctps0.shape[0]
    n_blocks = dctps0.shape[1]
    
    is_passed = np.zeros(n_ctps, dtype=np.bool_)
    
    # loop over all ctps
    for idx_ctp in range(n_ctps):
        # check if any cycle blocks this ctp
        for idx_cycle in range(n_cycles):
            dcpts0summed = 0
            for idx_block in range(n_blocks):
                dcpts0summed += winding_numbers[idx_cycle, idx_block] * dctps0[idx_ctp, idx_block]
            # if sum is not zero, current cycle blocks ctp
            if not dcpts0summed % cycle_lengths[idx_cycle] == 0:
                is_passed[idx_ctp] = False
                break
        else:
            # if loop does not break, ctp is passed
            is_passed[idx_ctp] = True
    
    return is_passed





@numba.njit()
def is_valid_nestcog_numba(windings, lengths, dctps0, n_ctps_wanted):
    """
    For a given nestcog phase cycle determine if it passes all wanted CTPs and
    blocks all unwanted CTPs.
    
    Parameters
    ----------
    windings : (n_cycles, n_blocks) array of int
        Winding numbers for each cycle.
    lengths : (n_cycles,) array of int
        Cycle lengths of each inner cycle. The number of scans is the product
        all numbers in this array.
    dctps0 : (n_ctps, n_blocks) array of int
        Coherence order changes relative to the coherence order changes of
        the reference path. The desired paths must be given first.
    n_ctps_wanted : int
        Number of desired paths. The first n_ctps_wanted rows of dctps0 must
        be the desired paths, the following ones the undesired ones.
    
    Returns
    -------
    is_valid : int
        Wether the phase cycle is valid or not. 0 indicates a valid cycle,
        1 a cycle that filters desired paths and 2 a cycle that does not
        filter desired paths but also not all undesired paths.
    """
    
    # check size and shape of input
    if not lengths.ndim == 1:
        raise ValueError('Internal error: `lengths` must be 1D.')
    if not windings.ndim == 2:
        raise ValueError('Internal error: `windings` must be 2D.')
    if not dctps0.ndim == 2:
        raise ValueError('Internal error: `dctps0` must be 2D.')
    if not lengths.shape[0] == windings.shape[0]:
        raise ValueError('Internal error: First dimensions of `lengths` '
                         'and `windings` must have same size.')
    if not windings.shape[1] == dctps0.shape[1]:
        raise ValueError('Internal error: Second dimensions of `windings` '
                         'and `dctps0` must have same size.')
    if n_ctps_wanted < 0 or n_ctps_wanted > dctps0.shape[0]:
        raise ValueError('Internal error: `n_ctps_wanted` must be between 0 '
                         'and number of given CTPs.')
        
    n_cycles = lengths.shape[0]
    n_ctps = dctps0.shape[0]
    n_blocks = dctps0.shape[1]
    if n_cycles < 1: return 3
    
    # loop over all wanted CTPs
    for idx_ctp in range(n_ctps_wanted):
        # check if any cycle blocks this CTP
        for idx_cycle in range(n_cycles):
            if lengths[idx_cycle] <= 1: continue
            dcpts0summed = 0
            for idx_block in range(n_blocks):
                dcpts0summed += windings[idx_cycle, idx_block] * dctps0[idx_ctp, idx_block]
            # if sum is not zero, current cycle blocks CTP
            if not dcpts0summed % lengths[idx_cycle] == 0:
                return 1
    
    # loop over all unwanted CTPs
    for idx_ctp in range(n_ctps_wanted, n_ctps):
        # check if any cycle blocks this CTP
        for idx_cycle in range(n_cycles):
            if lengths[idx_cycle] <= 1: continue
            dcpts0summed = 0
            for idx_block in range(n_blocks):
                dcpts0summed += windings[idx_cycle, idx_block] * dctps0[idx_ctp, idx_block]
            # if sum is not zero, current cycle blocks CTP
            if not dcpts0summed % lengths[idx_cycle] == 0:
                break
        else:
            # if loop does not break, CTP is passed
            return 2
    
    return 0





@numba.njit(parallel=True)
def are_valid_nestcogs_numba(windings, lengths, dctps0, n_ctps_wanted, n_check=0):
    """
    For several given nestcog phase cycles determine if they pass all wanted
    CTPs and blocks all unwanted CTPs.
    
    Parameters
    ----------
    lengths : (n_nestcogs, n_cycles,) array of int
        Cycle lengths of each inner cycle for every nestcog cycle.
        The number of scans is the product all numbers for a given nestcog
        cycle.
    windings : (n_nestcogs, n_cycles, n_blocks) array of int
        Winding numbers for each inner cycle and for each nestcog cycle.
    dctps0 : (n_ctps, n_blocks) array of int
        Coherence order changes relative to the coherence order changes of
        the reference path. The desired paths must be given first.
    n_ctps_wanted : int
        Number of desired paths. The first n_ctps_wanted rows of dctps0 must
        be the desired paths, the following ones the undesired ones.
    
    Returns
    -------
    is_valid : array of int
        Wether the phase cycle is valid or not for each nestcog cycle.
        0 indicates a valid cycle, 1 a cycle that filters desired paths and 2
        a cycle that does not filter desired paths but also not all undesired
        paths.
    """
    
    # check size and shape of input
    if not lengths.ndim == 2:
        raise ValueError('Internal error: `lengths` must be 2D.')
    if not windings.ndim == 3:
        raise ValueError('Internal error: `windings` must be 3D.')
    if not dctps0.ndim == 2:
        raise ValueError('Internal error: `dctps0` must be 2D.')
    
    n_cycles = windings.shape[0] # number of nestcog cycles
    n_subcyc = windings.shape[1] # number of subcycles per nestcog
    n_blocks = windings.shape[2] # number of pulse blocks


    if not lengths.shape[0] == n_cycles:
        raise ValueError('Internal error: First dimensions of `lengths` '
                         'and `windings` must have same size.')
    if not lengths.shape[1] == n_subcyc:
        raise ValueError('Internal error: First dimensions of `lengths` '
                         'and second of `windings` must have same size.')
    if not dctps0.shape[1] == n_blocks:
        raise ValueError('Internal error: Second dimensions of `dctps0` '
                         'and third of `windings` must have same size.')
    
    if n_ctps_wanted < 0 or n_ctps_wanted > dctps0.shape[0]:
        raise ValueError('Internal error: `n_ctps_wanted` must be between 0 '
                         'and number of given CTPs.')

    if n_check <= 0:
        # check every cycle
        n_check = n_cycles
    else:
        n_check = min(n_check, n_cycles)

    # allocate output array
    results = np.zeros(n_check, dtype=np.int64)
    results.fill(3)
    
    # parallel loop over all cycles
    for idx_cycle in numba.prange(n_check):
        
        # check validity of current cycle
        res = is_valid_nestcog_numba(
                windings[idx_cycle,:,:], lengths[idx_cycle,:],
                dctps0, n_ctps_wanted)
        results[idx_cycle] = res
    
    return results



def search_nestcog_exhaustive(n_scans_max, dctps0, n_ctps_wanted,
                            last_zero=False, no_inverse=True,
                            no_cfscan=True, no_cfwind=True,
                            n_find=1, n_scans_min=2, verbose=1):
    
    # check input
    n_scans_min = max(2, int(n_scans_min))
    n_scans_max = max(n_scans_min, int(n_scans_max))
    
    n_find = max(1, int(n_find))
    
    dctps0 = np.asarray(dctps0)
    if not dctps0.ndim == 2:
        raise ValueError("Must be 2D")
    
    n_ctps   = dctps0.shape[0]
    n_blocks = dctps0.shape[1]
    
    if n_ctps_wanted < 0 or n_ctps_wanted > n_ctps:
        raise ValueError("Must be inbetween")
        
    n_checked = 0
    
    max_MB_buffer = 32
    
    # print some info
    print("SEARCHING NESTCOG CYCLES EXHAUSTIVELY\n")
    print("Number of CTPs ", n_ctps)
    print("Number of pulse blocks: ", n_blocks)
    print("Probing number of scans from {} to {}".format(n_scans_min, n_scans_max))
    
    print("Maximum memory of buffers: {} MB".format(max_MB_buffer))
    
    print("\n  --> Starting the run <--")
    print("-"*30)
    
    for n_scans in range(n_scans_min, n_scans_max+1):
        print('  --> Number of scans: ', n_scans)
        found_solution = False
        
        factorizations = _generate_all_factorizations_numba(n_scans, 0)
        n_factors = factorizations.shape[1]
        
        # allocate buffers
        n_buffered = int(np.floor( max_MB_buffer * (1024**2/((1+n_blocks)*n_factors*factorizations.itemsize)) ))
        #n_buffered = 10000
        if n_buffered < 1:
            raise ValueError("Cannot allocate buffer.")
        buffer_windings = np.empty((n_buffered,n_factors,n_blocks), dtype=factorizations.dtype)
        buffer_lengths  = np.empty((n_buffered,n_factors), dtype=factorizations.dtype)
        buffer_windings[:,:,:] = 0
        buffer_lengths[:,:] = 0
        
        # allocate seeds
        seed_windings = np.empty((n_factors,n_blocks), dtype=factorizations.dtype)
        seed_lengths  = np.empty((n_factors,), dtype=factorizations.dtype)
        seed_windings[:,:] = 0
        seed_lengths[:]    = 0
        
        # allocate output arrays
        output_windings = np.zeros((n_find,n_factors,n_blocks), dtype=factorizations.dtype)
        output_lengths  = np.zeros((n_find,n_factors), dtype=factorizations.dtype)
        
        # initialize the seeds
        res = _init_nestwind_numba(seed_windings, seed_lengths, factorizations,
                        last_zero, no_inverse, no_cfscan, no_cfwind)
        if not res == 0:
            ValueError("Could not init")
            
        all_done = False
        n_found = 0
        while True:
            
            # fill buffer
            n_set, seed_windings, seed_lengths, all_done = _fill_buffer_nestcog_numba(
                buffer_windings, buffer_lengths, seed_windings, seed_lengths,
                factorizations, last_zero, no_inverse, no_cfscan, no_cfwind)
            
            if n_set < 0:
                raise ValueError("Could not fill buffer.")
            
            # check if there are any valid cycles
            results = are_valid_nestcogs_numba(
                buffer_windings, buffer_lengths, dctps0, n_ctps_wanted, n_set)
            
            # copy possible solutions to output array
            _       = _copy_by_mask_numba(buffer_windings, output_windings, results, n_found)
            n_found = _copy_by_mask_numba(buffer_lengths, output_lengths, results, n_found)
            
            # break
            if all_done or n_found == n_find:
                break
            
        if n_found > 0:
            print("FOUND SOLUTION")
            
            return output_windings[:n_found,:,:], output_lengths[:n_found,:]
        
        # free space of buffer
        del buffer_windings, buffer_lengths
        
    return output_windings[:n_found,:,:], output_lengths[:n_found,:]






# small script to find phase cycles
import warnings
import numpy as np
import numba as nb






#######################################
#                                     #
#   AAA   RRRR   RRRR    AAA   Y   Y  #
#  A   A  R   R  R   R  A   A   Y Y   #
#  AAAAA  RRRR   RRRR   AAAAA    Y    #
#  A   A  R  R   R  R   A   A    Y    #
#  A   A  R   R  R   R  A   A    Y    #
#                                     #
#######################################



@nb.njit(['b1(i4[:],i4[:,:])', 'b1(i8[:],i8[:,:])',
          'b1(u4[:],u4[:,:])', 'b1(u8[:],u8[:,:])'],
         cache=True)
def _row_in_array_numba(row, array):
    """Check wether a row is in a 2D array.
    
    Parameters
    ----------
    row : (m,) array_like
        The row to search for in the array.
    array : (m,n) array_like
        The array in which the given row is searched.
    
    Returns
    -------
    is_in_array : bool
        Wether the row is in the array or not.
    """
    
    # check shape of input
    if not row.ndim == 1:
        raise ValueError("`row` must be a 1D array.")
    if not array.ndim == 2:
        raise ValueError("`array` must be a 2D array.")
    
    m, n = array.shape[0], array.shape[1]

    if not row.size == n:
        raise ValueError("Number of elements in `row` does not match number of columns in `array`.")

    # loop over all rows in array
    for i in range(m):
        # loop over all entries in current row of array
        for j in range(n):
            if not array[i,j] == row[j]:
                break
        else:
            # if the loop for this row of array runs through, it equals given row
            return True
    else:
        return False





@nb.njit(['b1(i4[:],i4[:,:])', 'b1(i8[:],i8[:,:])',
          'b1(u4[:],u4[:,:])', 'b1(u8[:],u8[:,:])'],
         cache=True)
def _col_in_array_numba(col, array):
    """Check wether a column is in a 2D array.
    
    Parameters
    ----------
    col : (n,) array_like
        The column to search for in the array.
    array : (m,n) array_like
        The array in which the given column is searched.
    
    Returns
    -------
    is_in_array : bool
        Wether the column is in the array or not.
    """
    
    # check shape of input
    if not col.ndim == 1:
        raise ValueError("`col` must be a 1D array.")
    if not array.ndim == 2:
        raise ValueError("`array` must be a 2D array.")
    
    m, n = array.shape[0], array.shape[1]

    if not col.size == m:
        raise ValueError("Number of elements in `col` does not match number of rows in `array`.")
    
    # loop over all first cols in array
    for i in range(n):
        # loop over all entries in current col of array
        for j in range(m):
            if not array[j,i] == col[j]:
                break
        else:
            # if the loop for this col of array runs through, it equals given col
            return True
    else:
        return False





@nb.njit(cache=True)
def _delete_ones_numba(a):
    """Delete ones from a 1D array.
    
    Returns an array with the same elements as the given array except for
    ones, which are omitted.
    
    Parameters
    ----------
    a : 1D array_like
        Input array, which will not be altered by this function.
    
    Returns
    -------
    a_new : 1D array_like
        New array with ones deleted.
    """
    
    if not a.ndim == 1:
        raise ValueError("`a` must be 1D.")
    
    a_new = np.zeros_like(a)
    n_not_ones = 0
    for i in range(a.size):
        if not a[i] == 1:
            a_new[n_not_ones] = a[i]
            n_not_ones += 1
    
    return a_new[:n_not_ones]




@nb.njit(parallel=False)
def _copy_by_mask_numba(a_in, a_out, mask, key, idx_out):
    """Copy slices from an input array to an output array.

    Parameters
    ----------
    a_in : (n_in,...) array
        Input array to copy the slices from.
    a_out : (n_out,...) array
        Output array to copy the slices to.
    mask : (n_in,) array
        Key values that decide what elements are copied form the input array.
        Only if mask[i]==key, the i-th slices is copied.
    key : scalar
        Value that decides if element is copied or not.
    idx_out : int
        Index of the first slice in output array where a slices can be copied.
        This number must be between 0 and n_out.

    Returns
    -------
    idx_out : int
        New final index.
    """
    
    # check shape and size of input parameters
    if not mask.ndim == 1:
        raise ValueError("`mask` must be 1D.")
    if not a_in.ndim == a_out.ndim:
        raise ValueError("`a_in` and `a_out` must have the same number of dimensions.")
    for idx in range(1, a_out.ndim):
        if not a_in.shape[idx] == a_out.shape[idx]:
            raise ValueError("`a_in` and `a_out` must have the same shape (except first dimension).")
            
    size_in  = mask.size
    size_out = a_out.shape[0]
    
    if not a_in.shape[0] >= size_in:
        raise ValueError("Need more input slices.")
    if idx_out < 0 or idx_out > size_out:
        raise ValueError("`idx_out` is out of range.")
    
    # loop over all input slices
    for idx in range(size_in):
        
        # output is full
        if idx_out >= size_out:
            return size_out
        
        # copy input slice in output
        if mask[idx] == key:
            a_out[idx_out,...] = a_in[idx,...]
            idx_out += 1
    
    return idx_out






import math
import numpy as np
import numba as nb

# collection of cool algorithms compiled with numba

@nb.njit()
def check_1Darrays_equal_numba(a, b):
    """
    Check wether two 1D arrays are equal or not.

    Parameters
    ----------
    a : 1D array
        First array.
    b : 1D array
        Second array.
    
    Returns
    -------
    are_equal : bool
        Whether the arrays are equal or not.
    """

    # check input shape
    if a.ndim != 1 or b.ndim != 1:
        raise ValueError("`a` and `b` must be 1D arrays.")
    if a.size != b.size:
        raise ValueError("`a` and `b` must have same size.")

    for i in range(a.size):
        if a[i] != b[i]:
            return False

    return True





@nb.njit()
def check_2Darrays_equal_numba(a, b):
    """
    Check wether two 2D arrays are equal or not.

    Parameters
    ----------
    a : 2D array
        First array.
    b : 2D array
        Second array.
    
    Returns
    -------
    are_equal : bool
        Whether the arrays are equal or not.
    """

    # check input shape
    if a.ndim != 2 or  b.ndim != 2:
        raise ValueError("`a` and `b` must be 2D arrays.")
    if not a.shape[0] == b.shape[0]:
        raise ValueError("`a` and `b` must have same number of rows.")
    if not a.shape[1] == b.shape[1]:
        raise ValueError("`a` and `b` must have same number of columns.")
    
    n_rows = a.shape[0]
    n_cols = a.shape[1]

    for i in range(n_rows):
        for j in range(n_cols):
            if a[i,j] != b[i,j]:
                return False

    return True





@nb.njit()
def check_3Darrays_equal_numba(a, b):
    """
    Check wether two 3D arrays are equal or not.

    Parameters
    ----------
    a : 2D array
        First array.
    b : 2D array
        Second array.
    
    Returns
    -------
    are_equal : bool
        Whether the arrays are equal or not.
    """

    # check input shape
    if a.ndim != 3 or  b.ndim != 3:
        raise ValueError("`a` and `b` must be 3D arrays.")
    if not a.shape[0] == b.shape[0]:
        raise ValueError("`a` and `b` must have same number of tubes.")
    if not a.shape[1] == b.shape[1]:
        raise ValueError("`a` and `b` must have same number of rows.")
    if not a.shape[2] == b.shape[2]:
        raise ValueError("`a` and `b` must have same number of columns.")
    
    n_tubes = a.shape[0]
    n_rows  = a.shape[1]
    n_cols  = a.shape[2]

    for i in range(n_tubes):
        for j in range(n_rows):
            for k in range(n_cols):
                if a[i,j,k] != b[i,j,k]:
                    return False

    return True





@nb.njit()
def check_2Darray_all_rows_unique_numba(a):
    """
    Check if all rows are unique in a 2D array.
    
    Parameters
    ----------
    a : 2D array
        The array whose rows have to be checked for uniqueness.
    
    Returns
    -------
    all_rows_unique : bool
        Wether all rows are unique or not.
    """
    
    # check shape of input
    if not a.ndim == 2:
        raise ValueError("`a` must be a 2D array.")
    
    # get shape of input array
    n_rows = a.shape[0]
    n_cols = a.shape[1]
    
    # loop over all pairs of rows
    for i in range(n_rows):
        for j in range(i+1, n_rows):
            
            # check if rows are different
            for k in range(n_cols):
                if a[i,k] != a[j,k]:
                    break
            else:
                return False
    
    return True





@nb.njit(parallel=True)
def check_2Darrays_all_rows_unique_numba(arrays):
    """
    Check if all rows are unique for a list of 2D arrays.
    
    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        The arrays whose rows have to be checked for uniqueness.
    
    Returns
    -------
    all_rows_unique : (n_arrays,) array
        Wether all rows are unique or not for each 2D array.
    """
    
    # check shape of input
    if not arrays.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # get shape of input array and allocate output array
    n_arrays = arrays.shape[0]
    are_unique = np.zeros(n_arrays, dtype=np.bool_)
        
    # loop over all subarrays
    for idx in nb.prange(n_arrays):
        are_unique[idx] = check_2Darray_all_rows_unique_numba(arrays[idx])
    
    return are_unique





@nb.njit()
def sort_1Darray_numba(a):
    """In place sorting of a 1D array using bottom up mergesort.
    
    Parameters
    ----------
    a : 1D array
        Array to be sorted.
    
    Returns
    -------
    None
    """

    if not a.ndim == 1:
        raise ValueError("`a` must be 1D.")
    
    n = a.size
    working = np.zeros_like(a)
    
    # loop over all widths of subarrays 1, 2, 4, 8, ...
    width = 1
    while width < n:
        
        # loop over every consecutive pair of subarrays
        idx_start = 0
        while idx_start < n:
            
            # merge sorted subarrays to sorted array of twice the width
            idx_middle = min(idx_start+width, n)
            idx_end    = min(idx_start+width+width, n)
            idx_left  = idx_start
            idx_right = idx_middle
            
            k = idx_left
            while k < idx_end:
                if (idx_left < idx_middle 
                    and (idx_right >= idx_end or a[idx_left] <= a[idx_right])):
                    working[k] = a[idx_left]
                    idx_left += 1
                else:
                    working[k] = a[idx_right]
                    idx_right += 1
                k += 1
            
            # next pair of subarrays
            idx_start += width+width
        
        # copy higher sorted working array back to a
        a[:] = working[:]
        
        # double width for next iteration
        width *= 2





@nb.njit(parallel=True)
def arglexsort_3Darray_hslice_numba(a):
    """
    Sort.

    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        Array of 2D subarrays of shape (n_rows,n_cols) that have to
        be sorted.
    
    Returns
    -------
    argsort : (n_arrays,)
        TODO
    """
    
    # check shape of input
    if not a.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # Todo
    # get shape of input array
    n_arrays = a.shape[0]
    n_rows   = a.shape[1]
    n_cols   = a.shape[2]
    
    # allocate space for mergesort
    indices = np.zeros(n_arrays, dtype=np.int64)
    for i in range(indices.size):
        indices[i] = i
    working = np.zeros_like(indices)
    
    # loop over all widths of subarrays 1, 2, 4, 8, ...
    width = 1
    while width < n_arrays:
        
        # loop over every consecutive pair of subarrays
        
        for idx in nb.prange((n_arrays-1)//(2*width) + 1):
            
            idx_start = idx*(width+width)
            idx_middle = min(idx_start+width, n_arrays)
            idx_end    = min(idx_start+width+width, n_arrays)
            idx_left  = idx_start
            idx_right = idx_middle
            
            k = idx_left
            while k < idx_end:
                
                # check if we want to copy the left or right part element
                if idx_left < idx_middle:
                    if idx_right >= idx_end:
                        take_left = True
                    else:
                        # need to compare
                        left_not_bigger = True
                        done = False
                        a_idx_left  = indices[idx_left]
                        a_idx_right = indices[idx_right]
                        for idx_row in range(n_rows):
                            for idx_col in range(n_cols):
                                if a[a_idx_left,idx_row,idx_col] > a[a_idx_right,idx_row,idx_col]:
                                    left_not_bigger = False
                                    done = True
                                    break
                                elif a[a_idx_left,idx_row,idx_col] < a[a_idx_right,idx_row,idx_col]:
                                    done = True
                                    break
                            if done: break
                        take_left = left_not_bigger
                else:
                    take_left = False
            
                # copy the correct element
                if take_left:
                    working[k] = indices[idx_left]
                    idx_left += 1
                else:
                    working[k] = indices[idx_right]
                    idx_right += 1
                
                k += 1
        
        indices[:] = working[:]
        
        # double width for next iteration
        width = 2*width
        
    return indices





@nb.njit()
def check_3Darray_unique_hslice_numba(a):
    """
    For an array of 2D arrays, check which are unique and which are not.

    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        Array of 2D subarrays of shape (n_rows,n_cols) that have to
        be checked for uniqueness.

    Returns
    -------
    is_unique : (n_arrays) array of bool
        For every 2D subarray wether it is unique in `a` or not.
    """
    
    # check shape of input
    if not a.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # get shape of input array
    n_arrays = a.shape[0]
    n_rows   = a.shape[1]
    n_cols   = a.shape[2]
    
    # argsort array to speed up check for uniqueness
    argsort = arglexsort_3Darray_hslice_numba(a)
    # allocate output array
    is_unique = np.ones(n_arrays, dtype=np.bool_)
    
    # check all consecutive pairs of 2D subarrays
    for i in range(n_arrays-1):
        
        # compare both 2D arrays, if equal not unique
        are_equal = True
        for idx_row in range(n_rows):
            for idx_col in range(n_cols):
                if not a[argsort[i],idx_row,idx_col] == a[argsort[i+1],idx_row,idx_col]:
                    are_equal = False
                    break
            if not are_equal: break
        
        if are_equal:
            is_unique[argsort[i]]   = False
            is_unique[argsort[i+1]] = False
    
    return is_unique





@nb.njit()
def arglexsort_2Darray_rows_numba(a):
    """
    Argsort rows of a 2D array lexicographically. The original array is not 
    altered.

    Parameters
    ----------
    a : (n_rows,n_cols) 2D array
        Array whose columns have to be sorted. For the lexsort, the element in
        the first column of each row has highest priority.
    
    Returns
    -------
    argsort : (n_cols,)
        Indices that indicate the lexicographical order of columns. The sorted
        array is then a[argsort,:].
    """
    
    # check shape of input
    if not a.ndim == 2:
        raise ValueError("`a` must be a 2D array.")
    
    # get shape of input array
    n_rows   = a.shape[0]
    n_cols   = a.shape[1]
    
    # allocate space for mergesort
    indices = np.zeros(n_rows, dtype=np.int64)
    for i in range(indices.size):
        indices[i] = i
    working = np.zeros_like(indices)
    
    # loop over all widths of subarrays 1, 2, 4, 8, ...
    width = 1
    while width < n_rows:
        
        # loop over every consecutive pair of subarrays
        idx_start = 0
        while idx_start < n_rows:
            
            # merge sorted subarrays to sorted array of twice the width
            idx_middle = min(idx_start+width, n_rows)
            idx_end    = min(idx_start+width+width, n_rows)
            idx_left  = idx_start
            idx_right = idx_middle
            
            k = idx_left
            while k < idx_end:
                
                # check if we want to copy the left or right part element
                if idx_left < idx_middle:
                    if idx_right >= idx_end:
                        take_left = True
                    else:
                        # need to compare
                        left_not_bigger = True
                        a_idx_left  = indices[idx_left]
                        a_idx_right = indices[idx_right]
                        for idx_col in range(n_cols):
                            if a[a_idx_left,idx_col] > a[a_idx_right,idx_col]:
                                left_not_bigger = False
                                break
                            elif a[a_idx_left,idx_col] < a[a_idx_right,idx_col]:
                                break
                        take_left = left_not_bigger
                else:
                    take_left = False
            
                # copy the correct element
                if take_left:
                    working[k] = indices[idx_left]
                    idx_left += 1
                else:
                    working[k] = indices[idx_right]
                    idx_right += 1
                
                k += 1
            
            # next pair of subarrays
            idx_start += width+width
        
        indices[:] = working[:]
        
        # double width for next iteration
        width *= 2
        
    return indices





@nb.njit(parallel=True)
def lexsort_2Darrays_rows_numba(a):
    """
    Sort rows of a list of 2D arrays lexicographically in place.
    
    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        List of `n_arrays` 2D subarrays of shape (n_rows,n_cols) whose have
        to be sorted by columns. For the lexsort, the element in
        the first column of each row has highest priority.
    
    Returns
    -------
    None
    """
    
    # check shape of input
    if not a.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # get shape of input array
    n_arrays = a.shape[0]
    
    for idx_array in nb.prange(n_arrays):
        argsort = arglexsort_2Darray_rows_numba(a[idx_array])
        a[idx_array,:,:] = a[idx_array,argsort,:]





@nb.njit()
def get_3Darray_unique_hslice_indices_numba(a):
    """
    Get indices of unique 2D horizontal slices of a 3D array.
    
    Parameters
    ----------
    a : (n_tubes,n_rows,n_cols) 3D array
        Array to get the indices of unique slices (cuts along first dimension)
        of.
    
    Returns
    -------
    indices_unique : (<=n_slices,) 1D array
        Indices of the unique slices. The unique and sorted subset of slices is
        then a[indices_unique].
    """
    
    if not a.ndim == 3:
        raise ValueError("`a`must be a 3D array.")
        
    n_tubes = a.shape[0]
    n_rows   = a.shape[1]
    n_cols   = a.shape[2]
    
    
    indices_unique = np.zeros(n_tubes, dtype=np.int64)
    argsort = arglexsort_3Darray_hslice_numba(a)
    
    idx_pivot = argsort[0]
    indices_unique[0] = idx_pivot
    n_unique = 1
    for idx_slice in range(1, n_tubes):
        
        # compare this one with
        are_equal = True
        for idx_row in range(n_rows):
            for idx_col in range(n_cols):
                if not a[idx_pivot,idx_row,idx_col] == a[argsort[idx_slice],idx_row,idx_col]:
                    are_equal = False
                    break
            if not are_equal: break
        
        if not are_equal:
            indices_unique[n_unique] = argsort[idx_slice]
            idx_pivot = argsort[idx_slice]
            n_unique += 1
    
    return indices_unique[:n_unique]





@nb.njit()
def is_lexsorted_3Darray_hslice_numba(a):
    """
    Check wether slices along first dimension of a 3D array are
    lexicographically sorted.
    
    The importance of an entry decrases along a row and then along the columns.
    Hence a[:,0,0] is of highest importance followed by a[:,0,1] and so on.

    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        Array of 2D subarrays of shape (n_rows,n_cols) that have to be sorted.
    
    Returns
    -------
    is_sorted : bool
        Wether the arrays is lexicographically sorted.
    """
    
    # check shape of input
    if not a.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # get shape of input array
    n_arrays = a.shape[0]
    n_rows   = a.shape[1]
    n_cols   = a.shape[2]
    
    for i in range(1,n_arrays):
        
        left_not_bigger = False
        for idx_row in range(n_rows):
            for idx_col in range(n_cols):
                if a[i-1,idx_row,idx_col] < a[i,idx_row,idx_col]:
                    left_not_bigger = True
                    break
                elif a[i-1,idx_row,idx_col] > a[i,idx_row,idx_col]:
                    return False
            if left_not_bigger: break
    
    return True





@nb.njit()
def is_lexsorted_and_unique_3Darray_hslice_numba(a):
    """
    Check wether slices along first dimension of a 3D array are
    lexicographically sorted and all elements are unique.
    
    The importance of an entry decrases along a row and then along the columns.
    Hence a[:,0,0] is of highest importance followed by a[:,0,1] and so on.

    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        Array of 2D subarrays of shape (n_rows,n_cols) that have to be sorted.
    
    Returns
    -------
    is_sorted : bool
        Wether the arrays is lexicographically sorted and all slices are
        unqiue.
    """
    
    # check shape of input
    if not a.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # get shape of input array
    n_arrays = a.shape[0]
    n_rows   = a.shape[1]
    n_cols   = a.shape[2]
    
    for i in range(1,n_arrays):
        
        left_is_smaller = False
        left_equals_right = True
        for idx_row in range(n_rows):
            for idx_col in range(n_cols):
                if a[i-1,idx_row,idx_col] < a[i,idx_row,idx_col]:
                    left_is_smaller = True
                    left_equals_right = False
                    break
                elif a[i-1,idx_row,idx_col] > a[i,idx_row,idx_col]:
                    return False
            if left_is_smaller: break
        
        if left_equals_right:
            return False
    
    return True





@nb.njit()
def permute_3Darray_hslice_inplace_numba(array, permutation):
    """
    Rearranged the slices of a 3D array according to a permutation.
    
    The rearrangement is performed inplace so no additional memory has to be
    allocated.
    
    Parameters
    ----------
    array : (n,n_rows,n_cols) 3D array
        Array that has to be rearranged along the first dimension. The
        rearrangement will be perfomred inplace.
    permutation : (n,) array of int
        Array defining the permutation. `permutation[i] = j` means, that the
        slice `array[i]` will be where `array[j]` was after rearrangement.
        This array must contain every integer from 0 to n-1 exactly once. This
        array will also be changed inplace.
    
    Returns
    -------
    None
    """
    
    # check shape of input
    if not permutation.ndim == 1:
        raise ValueError("`permutation` must be a 1D array.")
    if not array.ndim == 3:
        raise ValueError("`array` must be a 3D array.")
    if not array.shape[0] == permutation.size:
        raise ValueError("First dimension of `array` must match size of `permutation`.")
    
    # get shape of input array
    n = permutation.size
    n_rows = array.shape[1]
    n_cols = array.shape[2]
    
    # loop over all starting positions
    for i in range(n):
        
        # if element already at correct position go to next position
        if permutation[i] == i:
            continue
        
        # have to do at maximum n-i swaps for this cycle
        for j in range(n-i-1):
            
            if permutation[i] <= i or permutation[i] >= n:
                raise ValueError("Index is out of allowed range. There is something wrong with the `permutation` array.")
            
            new_idx = permutation[i]
            
            # make swap
            for idx_row in range(n_rows):
                for idx_col in range(n_cols):
                    temp = array[new_idx,idx_row,idx_col]
                    array[new_idx,idx_row,idx_col] = array[i,idx_row,idx_col]
                    array[i,idx_row,idx_col] = temp
            permutation[i], permutation[new_idx] = permutation[new_idx], permutation[i]
            
            # if correct element arrived, we are done
            if permutation[i] == i:
                break
        
        else:
            raise ValueError("More swaps than necessary. There is something wrong with the `permutation` array.")





@nb.njit()
def lexsort_3Darray_hslice_numba(a):
    """
    Lexicographically sort the slices along first dimension of a 3D array.
    
    The importance of an entry decrases along a row and then along the columns.
    Hence a[:,0,0] is of highest importance followed by a[:,0,1] and so on.

    Parameters
    ----------
    a : (n_arrays,n_rows,n_cols) 3D array
        Array of 2D subarrays of shape (n_rows,n_cols) that have to be sorted.
    
    Returns
    -------
    None
    """
    
    # check shape of input
    if not a.ndim == 3:
        raise ValueError("`a` must be a 3D array.")
    
    # get shape of input array
    n_arrays = a.shape[0]
    n_rows   = a.shape[1]
    n_cols   = a.shape[2]
    
    # argsort array
    argsort = arglexsort_3Darray_hslice_numba(a)
    permutation = np.zeros(argsort.size, dtype=argsort.dtype)
    for i in range(permutation.size):
        permutation[argsort[i]] = i
    
    # apply permutation
    permute_3Darray_hslice_inplace_numba(a, permutation)





@nb.njit()
def are_hslices_shared_3Darrays_numba(a, b):
    """
    For two 3D arrays, whose slices are sorted and unique, find whether slices
    are shared between to arrays or not.
    
    Parameters
    ----------
    a : (n_a,n_rows,n_cols) 3D array
        First array.
    b : (n_b,n_rows,n_cols) 3D array
        Second array.
        
    Returns
    -------
    a_in_b : (n_a,) array of bool
        Wheter a slice of `a` appears in `b`
    b_in_a : (n_b,) array of bool
        Wheter a slice of `b` appears in `a`.
    """
    
    # check shape and size of input
    if (not a.ndim == 3) or (not b.ndim == 3):
        raise ValueError("`a` and `b` must be 3D.")
    if not a.shape[1] == b.shape[1]:
        raise ValueError("`a` and `b` must have the same number of rows.")
    if not a.shape[2] == b.shape[2]:
        raise ValueError("`a` and `b` must have the same number of columns.")
    
    # get shape of input array
    n_rows = a.shape[1]
    n_cols = a.shape[2]
    n_a    = a.shape[0]
    n_b    = b.shape[0]
    
    # check if elements are sorted and if all slices are unqique
    if not is_lexsorted_and_unique_3Darray_hslice_numba(a):
        raise ValueError("`a` is not sorted or not all slices are unqiue.")
    if not is_lexsorted_and_unique_3Darray_hslice_numba(b):
        raise ValueError("`b` is not sorted or not all slices are unqiue.")
    
    # allocate output arrays    
    a_in_b = np.zeros(shape=(n_a,), dtype=np.bool_)
    b_in_a = np.zeros(shape=(n_b,), dtype=np.bool_)
    
    # loop over all the elements of a and b
    idx_a = 0
    idx_b = 0
    while idx_a < n_a and idx_b < n_b:
        
        # determine wether elements are equal or not
        a_equals_b = True
        for idx_row in range(n_rows):
            for idx_col in range(n_cols):
                if a[idx_a,idx_row,idx_col] < b[idx_b,idx_row,idx_col]:
                    a_is_smaller = True
                    a_equals_b = False
                    break
                elif a[idx_a,idx_row,idx_col] > b[idx_b,idx_row,idx_col]:
                    a_is_smaller = False
                    a_equals_b = False
                    break
            if not a_equals_b: break
        
        if a_equals_b:
            a_in_b[idx_a] = True
            b_in_a[idx_b] = True
            idx_a += 1
            idx_b += 1
        elif a_is_smaller:
            idx_a += 1
        else:
            idx_b += 1
    
    return a_in_b, b_in_a


